import ast
import logging
import warnings
from collections.abc import Mapping, Sequence
from numbers import Number
from typing import Literal

import networkx as nx
import numpy as np
import pandas as pd
import xarray as xr

from .dataset import Dataset, construct
from .tree_branch import CachedTree, DataTreeBranch

try:
    from dask.array import Array as dask_array_type
except ModuleNotFoundError:
    dask_array_type = ()

try:
    from sparse import SparseArray as sparse_array_type
except ModuleNotFoundError:
    sparse_array_type = ()

try:
    from ast import unparse
except ImportError:
    from astunparse import unparse as _unparse

    def unparse(*args):
        return _unparse(*args).strip("\n")


logger = logging.getLogger("sharrow")

well_known_names = {
    "nb",
    "np",
    "pd",
    "xr",
    "pa",
    "log",
    "exp",
    "log1p",
    "expm1",
    "max",
    "min",
    "piece",
    "hard_sigmoid",
    "transpose_leading",
    "clip",
}

NOTSET = "<--NOTSET-->"


def _require_string(x):
    if not isinstance(x, str):
        raise ValueError("must be string")
    return x


def _iat(source, *, _names=None, _load=False, _index_name=None, **idxs):
    loaders = {}
    inum = 0

    def _ixname():
        if _index_name is not None:
            return _index_name
        nonlocal inum
        inum += 1
        return f"index{inum}"

    for k, v in idxs.items():
        if isinstance(v, xr.DataArray):
            loaders[k] = v
        else:
            loaders[k] = xr.DataArray(v, dims=[_ixname() for n in range(v.ndim)])
    if _names:
        ds = source[_names]
    else:
        ds = source
    if _load:
        ds = ds._load()
    return ds.isel(**loaders)


def _at(source, *, _names=None, _load=False, _index_name=None, **idxs):
    loaders = {}
    inum = 0

    def _ixname():
        if _index_name is not None:
            return _index_name
        nonlocal inum
        inum += 1
        return f"index{inum}"

    for k, v in idxs.items():
        if isinstance(v, xr.DataArray):
            loaders[k] = v
        else:
            loaders[k] = xr.DataArray(v, dims=[_ixname() for n in range(v.ndim)])
    if _names:
        ds = source[_names]
    else:
        ds = source
    if _load:
        ds = ds._load()
    return ds.sel(**loaders)


def gather(source, indexes):
    """
    Extract values by label on the coordinates indicated by columns of a DataFrame.

    Parameters
    ----------
    source : xarray.DataArray or xarray.Dataset
        The source of the values to extract.
    indexes : Mapping[str, array-like]
        The keys of `indexes` (if given as a dataframe, the column names)
        should match the named dimensions of `source`.  The resulting extracted
        data will have a shape one row per row of `df`, and columns matching
        the data variables in `source`, and each value is looked up by the labels.

    Returns
    -------
    pd.DataFrame
    """
    result = _at(source, **indexes).reset_coords(drop=True)
    return result


def igather(source, positions):
    """
    Extract values by position on the coordinates indicated by columns of a DataFrame.

    Parameters
    ----------
    source : xarray.DataArray or xarray.Dataset
    positions : pd.DataFrame or Mapping[str, array-like]
        The columns (or keys) of `df` should match the named dimensions of
        this Dataset.  The resulting extracted DataFrame will have one row
        per row of `df`, columns matching the data variables in this dataset,
        and each value is looked up by the positions.

    Returns
    -------
    pd.DataFrame
    """
    result = _iat(source, **positions).reset_coords(drop=True)
    return result


def xgather(source, positions, indexes):
    if len(indexes) == 0:
        return igather(source, positions)
    elif len(positions) == 0:
        return gather(source, indexes)
    else:
        return gather(igather(source, positions), indexes)


def _dataarray_to_numpy(self) -> np.ndarray:
    """Coerces wrapped data to numpy and returns a numpy.ndarray."""
    data = self.data
    if isinstance(data, dask_array_type):
        data = data.compute()
    if isinstance(data, sparse_array_type):
        data = data.todense()
    data = np.asarray(data)
    return data


class Relationship:
    """Defines a linkage between datasets in a `DataTree`."""

    def __init__(
        self,
        parent_data,
        parent_name,
        child_data,
        child_name,
        indexing="label",
        analog=None,
    ):
        self.parent_data = _require_string(parent_data)
        """str: Name of the parent dataset."""

        self.parent_name = _require_string(parent_name)
        """str: Variable in the parent dataset that references the child dimension."""

        self.child_data = _require_string(child_data)
        """str: Name of the child dataset."""

        self.child_name = _require_string(child_name)
        """str: Dimension in the child dataset that is used by this relationship."""

        if indexing not in {"label", "position"}:
            raise ValueError("indexing must be by label or position")
        self.indexing = indexing
        """str: How the target dimension is used, either by 'label' or 'position'."""

        self.analog = analog
        """str: Original variable that defined label-based relationship before digitization."""

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            if self.analog:
                left = (
                    f"<Relationship by label: "
                    f"{self.parent_data}[{self.analog!r}] -> "
                    f"{self.child_data}[{self.child_name!r}]>"
                )
            else:
                left = repr(self)
            if other.analog:
                right = (
                    f"<Relationship by label: "
                    f"{other.parent_data}[{other.analog!r}] -> "
                    f"{other.child_data}[{other.child_name!r}]>"
                )
            else:
                right = repr(other)
            return left == right

    def __repr__(self):
        return (
            f"<Relationship by {self.indexing}: "
            f"{self.parent_data}[{self.parent_name!r}] -> "
            f"{self.child_data}[{self.child_name!r}]>"
        )

    def attrs(self):
        return dict(
            parent_name=self.parent_name,
            child_name=self.child_name,
            indexing=self.indexing,
        )

    def to_dict(self):
        return dict(
            parent_data=self.parent_data,
            parent_name=self.parent_name,
            child_data=self.child_data,
            child_name=self.child_name,
            indexing=self.indexing,
            analog=self.analog,
        )

    @classmethod
    def from_string(cls, s):
        """
        Construct a `Relationship` from a string.

        Parameters
        ----------
        s : str
            The relationship definition.
            To create a label-based relationship, the string should look like
            "ParentNode.variable_name @ ChildNode.dimension_name".  To create
            a position-based relationship, give
            "ParentNode.variable_name -> ChildNode.dimension_name".

        Returns
        -------
        Relationship
        """
        if "->" in s:
            parent, child = s.split("->", 1)
            i = "position"
        elif "@" in s:
            parent, child = s.split("@", 1)
            i = "label"
        else:
            raise ValueError(f"cannot interpret relationship {s!r}")
        p1, p2 = parent.split(".", 1)
        c1, c2 = child.split(".", 1)
        p1 = p1.strip()
        p2 = p2.strip()
        c1 = c1.strip()
        c2 = c2.strip()
        return cls(
            parent_data=p1,
            parent_name=p2,
            child_data=c1,
            child_name=c2,
            indexing=i,
        )


class DataTree:
    """
    A tree representing linked datasets, from which data can flow.

    Parameters
    ----------
    graph : networkx.MultiDiGraph
    root_node_name : str or False
        The name of the node at the root of the tree.
    extra_funcs : Tuple[Callable]
        Additional functions that can be called by Flow objects created
        using this DataTree.  These functions should have defined `__name__`
        attributes, so they can be called in expressions.
    extra_vars : Mapping[str,Any], optional
        Additional named constants that can be referenced by expressions in
        Flow objects created using this DataTree.
    cache_dir : Path-like, optional
        The default directory where Flow objects are created.
    relationships : Iterable[str or Relationship]
        The relationship definitions used to define this tree.  All dataset
        nodes named in these relationships should also be included as
        keyword arguments for this constructor.
    force_digitization : bool, default False
        Whether to automatically digitize all relationships (converting them
        from label-based to position-based).  Digitization is required to
        evaluate Flows, but doing so automatically on construction may be
        inefficient.
    dim_order : Tuple[str], optional
        The order of dimensions to use in Flow outputs.  Generally only needed
        if there are multiple dimensions in the root dataset.
    aux_vars : Mapping[str,Any], optional
        Additional named arrays or numba-typable variables that can be
        referenced by expressions in Flow objects created using this DataTree.
    """

    DatasetType = Dataset

    def __init__(
        self,
        graph=None,
        root_node_name=None,
        extra_funcs=(),
        extra_vars=None,
        cache_dir=None,
        relationships=(),
        force_digitization=False,
        dim_order=None,
        aux_vars=None,
        **kwargs,
    ):
        if isinstance(graph, Dataset):
            raise ValueError("datasets must be given as keyword arguments")

        # raw init
        if graph is None:
            graph = nx.MultiDiGraph()
        self._graph = graph
        self._root_node_name = None
        self.force_digitization = force_digitization
        self.dim_order = dim_order
        self.dim_exclude = set()

        # defined init
        if root_node_name is not None and root_node_name in kwargs:
            self.add_dataset(root_node_name, kwargs[root_node_name])
        self.root_node_name = root_node_name
        self.extra_funcs = extra_funcs
        self.extra_vars = extra_vars or {}
        self.aux_vars = aux_vars or {}
        self.cache_dir = cache_dir
        self._cached_indexes = {}
        for k, v in kwargs.items():
            if root_node_name is not None and k == root_node_name:
                continue
            self.add_dataset(k, v)
        for r in relationships:
            self.add_relationship(r)
        if force_digitization:
            self.digitize_relationships(inplace=True)

        # These filters are applied to incoming datasets when using `replace_datasets`.
        self.replacement_filters = {}
        """Dict[Str,Callable]: Filters that are automatically applied to data on replacement.

        When individual datasets are replaced in the tree, the incoming dataset is
        passed through the filter with a matching name-key (if it exists).  The filter
        should be a function that accepts one argument (the incoming dataset) and returns
        one value (the dataset to save in the tree).  These filters can be used to ensure
        data quality, e.g. renaming variables, ensuring particular data types, etc.
        """

        self.subspace_fallbacks = {}
        """Dict[Str:List[Str]]: Allowable fallback subspace lookups.

        When a named variable is not found in a given subspace, the default result is
        raising a KeyError. But, if fallbacks are defined for a given subspace, the
        fallbacks are searched in order for the desired variable.
        """

    @property
    def shape(self):
        """Tuple[int]: base shape of arrays that will be loaded when using this DataTree."""
        if self.dim_order:
            dim_order = self.dim_order
        else:
            from .flows import presorted

            dim_order = presorted(self.root_dataset.dims, self.dim_order)
        return tuple(
            self.root_dataset.dims[i] for i in dim_order if i not in self.dim_exclude
        )

    @property
    def root_dims(self):
        from .flows import presorted

        return tuple(
            presorted(self.root_dataset.dims, self.dim_order, self.dim_exclude)
        )

    def __shallow_copy_extras(self):
        return dict(
            extra_funcs=self.extra_funcs,
            extra_vars=self.extra_vars,
            aux_vars=self.aux_vars,
            cache_dir=self.cache_dir,
            force_digitization=self.force_digitization,
            dim_order=self.dim_order,
        )

    def __repr__(self):
        s = f"<{self.__module__}.{self.__class__.__name__}>"
        if len(self._graph.nodes):
            s += "\n datasets:"
            if self.root_node_name:
                s += f"\n - {self.root_node_name}"
            for k in self._graph.nodes:
                if k == self.root_node_name:
                    continue
                s += f"\n - {k}"
        else:
            s += "\n datasets: none"
        if len(self._graph.edges):
            s += "\n relationships:"
            for e in self._graph.edges:
                s += f"\n - {self._get_relationship(e)!r}".replace(
                    "<Relationship ", ""
                ).rstrip(">")
        else:
            s += "\n relationships: none"
        return s

    def view_relationships(self, fontname="Arial", fontsize=9):
        from .viz import display_svg, make_graph

        return display_svg(make_graph(self, fontname=fontname, fontsize=fontsize))

    def _hash_features(self):
        h = []
        if len(self._graph.nodes):
            if self.root_node_name:
                h.append(f"dataset:{self.root_node_name}")
            for k in self._graph.nodes:
                if k == self.root_node_name:
                    continue
                h.append(f"dataset:{k}")
        else:
            h.append("datasets:none")
        if len(self._graph.edges):
            for e in self._graph.edges:
                r = f"relationship:{self._get_relationship(e)!r}".replace(
                    "<Relationship ", ""
                ).rstrip(">")
                h.append(r)
        else:
            h.append("relationships:none")
        h.append(f"dim_order:{self.dim_order}")
        return h

    @property
    def root_node_name(self):
        """str: The root node for this data tree, which is only ever a parent."""
        if self._root_node_name is None:
            for nodename in self._graph.nodes:
                if self._graph.in_degree(nodename) == 0:
                    self._root_node_name = nodename
                    break
        return self._root_node_name

    @root_node_name.setter
    def root_node_name(self, name):
        if name is None or name is False:
            self._root_node_name = name
            return
        if not isinstance(name, str):
            raise TypeError(
                f"root_node_name must be one of [str, None, False] not {type(name)}"
            )
        if name not in self._graph.nodes:
            raise KeyError(name)
        self._root_node_name = name

    @property
    def root_node_name_str(self):
        """str: The root node for this data tree, which is only ever a parent.

        This method raises a ValueError if root node cannot be determined.
        """
        if self._root_node_name is None:
            for nodename in self._graph.nodes:
                if self._graph.in_degree(nodename) == 0:
                    self._root_node_name = nodename
                    break
        if self._root_node_name is None:
            raise ValueError("root node cannot be determined")
        if self._root_node_name is False:
            raise ValueError("root node is False")
        return self._root_node_name

    def add_relationship(self, *args, **kwargs):
        """
        Add a relationship to this DataTree.

        The new relationship will point from a variable in one dataset
        to a dimension of another dataset in this tree.  Both the parent
        and the child datasets should already have been added.

        Parameters
        ----------
        *args, **kwargs
            All arguments are passed through to the `Relationship`
            contructor, unless only a single `str` argument is provided,
            in which case the `Relationship.from_string` class constructor
            is used.
        """
        if len(args) == 1 and isinstance(args[0], Relationship):
            r = args[0]
        elif len(args) == 1 and isinstance(args[0], str):
            r = Relationship.from_string(args[0])
        else:
            r = Relationship(*args, **kwargs)

        # check for existing relationships, don't duplicate
        for e in self._graph.edges:
            r2 = self._get_relationship(e)
            if r == r2:
                return

        # confirm correct pointer
        r.parent_data = self.finditem(r.parent_name, maybe_in=r.parent_data)

        self._graph.add_edge(r.parent_data, r.child_data, **r.attrs())
        if self.force_digitization:
            self.digitize_relationships(inplace=True)

    def get_relationship(self, parent, child):
        attrs = self._graph.edges[parent, child]
        return Relationship(parent_data=parent, child_data=child, **attrs)

    def list_relationships(self) -> list[Relationship]:
        """List : List all relationships defined in this tree."""
        result = []
        for e in self._graph.edges:
            result.append(self._get_relationship(e))
        return result

    def add_dataset(self, name, dataset, relationships=(), as_root=False):
        """
        Add a new Dataset node to this DataTree.

        Parameters
        ----------
        name : str
        dataset : Dataset or pandas.DataFrame
            Will be coerced into a `Dataset` object if it is not already
            in that format, using a no-copy process if possible.
        relationships : Tuple[str or Relationship]
            Also add these relationships.
        as_root : bool, default False
            Set this new node as the root of the tree, displacing any existing
            root.
        """
        self._graph.add_node(name, dataset=construct(dataset))
        if self.root_node_name is None or as_root:
            self.root_node_name = name
        if isinstance(relationships, str):
            relationships = [relationships]
        for r in relationships:
            # TODO validate relationships before adding.
            self.add_relationship(r)
        if self.force_digitization:
            self.digitize_relationships(inplace=True)

    def add_items(self, items):
        if isinstance(items, Sequence):
            for i in items:
                self.add_items(i)
        elif isinstance(items, Mapping):
            if "name" in items and "dataset" in items:
                self.add_dataset(items["name"], items["dataset"])
                preload = True
            else:
                preload = False
            for k, v in items.items():
                if preload and k in {"name", "dataset"}:
                    continue
                if k == "relationships":
                    for r in v:
                        self.add_relationship(r)
                else:
                    self.add_dataset(k, v)
        else:
            raise ValueError("add_items requires Sequence or Mapping")

    @property
    def root_node(self):
        return self._graph.nodes[self.root_node_name_str]

    @property
    def root_dataset(self):
        return self._graph.nodes[self.root_node_name_str]["dataset"]

    @root_dataset.setter
    def root_dataset(self, x):
        from .dataset import Dataset

        if not isinstance(x, Dataset):
            x = construct(x)
        if self.root_node_name_str in self.replacement_filters:
            x = self.replacement_filters[self.root_node_name](x)
        self._graph.nodes[self.root_node_name]["dataset"] = x

    def _get_relationship(self, edge):
        return Relationship(
            parent_data=edge[0], child_data=edge[1], **self._graph.edges[edge]
        )

    def __getitem__(self, item):
        if hasattr(self, "_eval_cache") and item in self._eval_cache:
            return self._eval_cache[item]
        try:
            return self.get(item)
        except KeyError as err:
            s = self._graph.nodes.get(item, {}).get("dataset", None)
            if s is not None:
                return DataTreeBranch(self, item)
            raise err

    def get(self, item, default=None, broadcast=True, coords=True):
        """
        Access variable(s) from this tree.

        Parameters
        ----------
        item : str or Sequence[str]
            Each value can be just the name of the variable if that name is unique
            within the tree, or use dotted notation ('node_name.var_name') to give
            the node name explicitly and resolve ambiguity as necessary.
        default
            If provided, this default value is used for any missing item(s).
        broadcast : bool, default True
            Broadcast all arrays up to the dimensions of the root node in the tree.
        coords : bool, default True
            Attach coordinates from the root node of the tree to the result.

        Returns
        -------
        DataArray or Dataset
        """
        if isinstance(item, (list, tuple)):
            from .dataset import Dataset

            return Dataset(
                {
                    k: self.get(k, default=default, broadcast=broadcast, coords=coords)
                    for k in item
                }
            )
        try:
            result = self._getitem(item, dim_names_from_top=True)
        except KeyError:
            try:
                result = self._getitem(
                    item, include_blank_dims=True, dim_names_from_top=True
                )
            except KeyError:
                if default is None:
                    raise
                else:
                    result = xr.DataArray(default)
        if self.root_node_name:
            root_dataset = self.root_dataset
            if result.dims != self.root_dims and broadcast:
                result, _ = xr.broadcast(result, root_dataset)
            if coords:
                add_coords = {}
                for i in result.dims:
                    if i not in result.coords and i in root_dataset.coords:
                        add_coords[i] = root_dataset.coords[i]
                if add_coords:
                    result = result.assign_coords(add_coords)
        elif self.root_node_name is False:
            if "." in item:
                item_in, item = item.split(".", 1)
                base_dataset = self._graph.nodes[item_in]["dataset"]
                if coords:
                    add_coords = {}
                    for i in result.dims:
                        if i not in result.coords and i in base_dataset.coords:
                            add_coords[i] = base_dataset.coords[i]
                    if add_coords:
                        result = result.assign_coords(add_coords)
        if broadcast:
            if self.dim_order is None:
                result = result.transpose(*self.root_dims)
            else:
                result = result.transpose(*self.dim_order)
        return result

    def finditem(self, item, maybe_in=None):
        if maybe_in is not None and maybe_in in self._graph.nodes:
            dataset = self._graph.nodes[maybe_in].get("dataset", {})
            if item in dataset:
                return maybe_in
        return self._getitem(item, just_node_name=True)

    def _getitem(
        self,
        item,
        include_blank_dims=False,
        only_dims=False,
        just_node_name=False,
        dim_names_from_top=False,
    ):
        if isinstance(item, (list, tuple)):
            from .dataset import Dataset

            return Dataset({k: self[k] for k in item})

        if "." in item:
            item_in, item = item.split(".", 1)
            queue = [self.root_node_name]
            if self.root_node_name is False:
                # when root_node_name is False, we don't want to broadcast
                # back to the root, but instead only to the given `item_in`
                queue = [item_in]
                item_in = None
        else:
            item_in = None
            queue = [self.root_node_name_str]
        examined = set()
        start_from = queue[0]
        while len(queue):
            current_node = queue.pop(0)
            if current_node in examined:
                continue
            dataset = self._graph.nodes[current_node].get("dataset", {})
            try:
                by_name = item in dataset and not only_dims
            except TypeError:
                by_name = False
            try:
                by_dims = not by_name and include_blank_dims and (item in dataset.dims)
            except TypeError:
                by_dims = False
            if (by_name or by_dims) and (item_in is None or item_in == current_node):
                if just_node_name:
                    return current_node
                if current_node == start_from:
                    if by_dims:
                        return xr.DataArray(
                            pd.RangeIndex(dataset.dims[item]), dims=item
                        )
                    else:
                        return dataset[item]
                else:
                    _positions = {}
                    _labels = {}
                    if by_dims:
                        if item in dataset.variables:
                            coords = {item: dataset.variables[item]}
                        else:
                            coords = None
                        result = xr.DataArray(
                            pd.RangeIndex(dataset.dims[item]),
                            dims=item,
                            coords=coords,
                        )
                    else:
                        result = dataset[item]
                    dims_in_result = set(result.dims)
                    top_dim_names = {}
                    for path in nx.algorithms.simple_paths.all_simple_edge_paths(
                        self._graph, start_from, current_node
                    ):
                        if dim_names_from_top:
                            e = path[0]
                            top_dim_name = self._graph.edges[e].get("parent_name")
                            start_dataset = self._graph.nodes[start_from]["dataset"]
                            # deconvert digitized dim names back to native dims
                            if (
                                top_dim_name not in start_dataset.dims
                                and top_dim_name in start_dataset.variables
                            ):
                                if start_dataset.variables[top_dim_name].ndim == 1:
                                    top_dim_name = start_dataset.variables[
                                        top_dim_name
                                    ].dims[0]
                        else:
                            top_dim_name = None
                        path_dim = self._graph.edges[path[-1]].get("child_name")
                        if path_dim not in dims_in_result:
                            continue
                        # path_indexing = self._graph.edges[path[-1]].get('indexing')
                        t1 = None
                        # intermediate nodes on path
                        for e, e_next in zip(path[:-1], path[1:]):
                            r = self._get_relationship(e)
                            r_next = self._get_relationship(e_next)
                            if t1 is None:
                                t1 = self._graph.nodes[r.parent_data].get("dataset")
                            t2 = self._graph.nodes[r.child_data].get("dataset")[
                                [r_next.parent_name]
                            ]
                            if r.indexing == "label":
                                t1 = t2.sel(
                                    {
                                        r.child_name: _dataarray_to_numpy(
                                            t1[r.parent_name]
                                        )
                                    }
                                )
                            else:  # by position
                                t1 = t2.isel(
                                    {
                                        r.child_name: _dataarray_to_numpy(
                                            t1[r.parent_name]
                                        )
                                    }
                                )
                        # final node in path
                        e = path[-1]
                        r = Relationship(
                            parent_data=e[0], child_data=e[1], **self._graph.edges[e]
                        )
                        if t1 is None:
                            t1 = self._graph.nodes[r.parent_data].get("dataset")
                        if r.indexing == "label":
                            _labels[r.child_name] = _dataarray_to_numpy(
                                t1[r.parent_name]
                            )
                        else:  # by position
                            _idx = _dataarray_to_numpy(t1[r.parent_name])
                            if not np.issubdtype(_idx.dtype, np.integer):
                                _idx = _idx.astype(np.int64)
                            _positions[r.child_name] = _idx
                        if top_dim_name is not None:
                            top_dim_names[r.child_name] = top_dim_name
                    if len(top_dim_names) > 1:
                        if len(set(top_dim_names.values())) == 1:
                            # capture the situation where all top dims are the same
                            _positions = {
                                k: xr.DataArray(v, dims=[top_dim_names[k]])
                                for (k, v) in _positions.items()
                            }
                            _labels = {
                                k: xr.DataArray(v, dims=[top_dim_names[k]])
                                for (k, v) in _labels.items()
                            }
                            # the top dim names have served their purpose, so clear them
                            top_dim_names = {}
                        elif len(set(top_dim_names.values())) < len(top_dim_names):
                            # capture the situation where some but not all top dims are the same
                            # same as above?
                            _positions = {
                                k: xr.DataArray(v, dims=[top_dim_names[k]])
                                for (k, v) in _positions.items()
                            }
                            _labels = {
                                k: xr.DataArray(v, dims=[top_dim_names[k]])
                                for (k, v) in _labels.items()
                            }
                            # the top dim names have served their purpose, so clear them
                            top_dim_names = {}
                    y = xgather(result, _positions, _labels)
                    if len(result.dims) == 1 and len(y.dims) == 1:
                        y = y.rename({y.dims[0]: result.dims[0]})
                    elif len(dims_in_result) == len(y.dims):
                        y = y.rename({_i: _j for _i, _j in zip(y.dims, result.dims)})
                    if top_dim_names:
                        y = y.rename(top_dim_names)
                    return y
            else:
                examined.add(current_node)
                for _, next_up in self._graph.out_edges(current_node):
                    if next_up not in examined:
                        queue.append(next_up)

        raise KeyError(item)

    def get_expr(
        self,
        expression,
        engine="sharrow",
        allow_native=True,
        *,
        dtype="float32",
        with_coords: bool = True,
        parser: Literal["pandas", "python"] = "pandas",
    ):
        """
        Access or evaluate an expression.

        Parameters
        ----------
        expression : str
        engine : {'sharrow', 'numexpr', 'python', 'pandas-numexpr'}
            The engine used to resolve expressions.  The numexpr engine uses
            that library directly, while the pandas-numexpr engine uses the
            pandas `eval` method with the numexpr engine.
        allow_native : bool, default True
            If the expression is an array in a dataset of this tree, return
            that array directly.  Set to false to force evaluation, which
            will also ensure proper broadcasting consistent with this data tree.
        dtype : str or dtype, default 'float32'
            The dtype to use when creating new arrays.  This only applies when
            the expression is not returned as a native variable from the tree.
        with_coords : bool, default True
            Attach coordinates from the root node of the tree to the result.
            If the coordinates are not needed in the result, the process
            of attaching them can be skipped.
        parser : {'pandas', 'python'}
            The parser to use when evaluating the expression. This argument
            only applies to pandas-based engines ('python' and 'pandas-numexpr').
            It is ignored when using the 'sharrow' or 'numexpr' engines.

        Returns
        -------
        DataArray
        """
        if np.issubdtype(dtype, np.number) and isinstance(dtype, type):
            dtype = dtype.__name__
        elif dtype is bool:
            dtype = "bool"
        try:
            if allow_native:
                result = self[expression]
            else:
                raise KeyError
        except (KeyError, IndexError):
            if engine == "sharrow":
                if dtype is None:
                    dtype = "float32"
                result = (
                    self.setup_flow({expression: expression}, dtype=dtype)
                    .load_dataarray()
                    .isel(expressions=0)
                )
            elif engine == "numexpr":
                import numexpr as ne
                from xarray import DataArray

                try:
                    result = DataArray(
                        ne.evaluate(expression, local_dict=CachedTree(self)),
                    )
                except Exception:
                    if dtype is None:
                        dtype = "float32"
                    result = (
                        self.setup_flow({expression: expression}, dtype=dtype)
                        .load_dataarray()
                        .isel(expressions=0)
                    )
                else:
                    if dtype is not None:
                        result = result.astype(dtype)
                    # numexpr doesn't carry over the dimension names or coords
                    result = result.rename(
                        {result.dims[i]: self.root_dims[i] for i in range(result.ndim)}
                    )
                    if with_coords:
                        result = result.assign_coords(self.root_dataset.coords)

            elif engine == "pandas-numexpr":
                from xarray import DataArray

                self._eval_cache = {}
                try:
                    result = DataArray(
                        pd.eval(
                            expression,
                            resolvers=[self],
                            engine="numexpr",
                            parser=parser,
                        ),
                    ).astype(dtype)
                except NotImplementedError:
                    result = DataArray(
                        pd.eval(
                            expression, resolvers=[self], engine="python", parser=parser
                        ),
                    ).astype(dtype)
                else:
                    # numexpr doesn't carry over the dimension names or coords
                    result = result.rename(
                        {result.dims[i]: self.root_dims[i] for i in range(result.ndim)}
                    )
                    if with_coords:
                        result = result.assign_coords(self.root_dataset.coords)
                finally:
                    del self._eval_cache
            elif engine == "python":
                from xarray import DataArray

                self._eval_cache = {}
                try:
                    result = DataArray(
                        pd.eval(
                            expression, resolvers=[self], engine="python", parser=parser
                        ),
                    ).astype(dtype)
                finally:
                    del self._eval_cache
            else:
                raise ValueError(f"unknown engine {engine}") from None
        return result

    def eval(
        self,
        expression: str | Number,
        engine: Literal[None, "numexpr", "sharrow", "python"] = None,
        *,
        dtype: np.dtype | str | None = None,
        name: str | None = None,
        with_coords: bool = True,
    ):
        """
        Evaluate an expression.

        The resulting DataArray will have dimensions that match the root
        Dataset of this tree, and the content will be broadcast to those
        dimensions if necessary.  The expression evaluated will be assigned
        as a scalar coordinate named 'expressions', to facilitate concatenation
        with other `eval` results if desired.

        Parameters
        ----------
        expression : str | Number
        engine : {None, 'numexpr', 'sharrow', 'python'}
            The engine used to resolve expressions. If None, the default is
            to try 'numexpr' first, then 'sharrow' if that fails.
        dtype : str or dtype, optional
            The dtype to use for the result.  If the engine is `sharrow` and
            no value is given, this will default to `float32`, otherwise the
            default is to use the dtype of the result of the expression.
        name : str, optional
            The name to give the resulting DataArray.

        Returns
        -------
        DataArray
        """
        # when passing in a numeric value or boolean, simply broadcast it to the root dims
        if isinstance(expression, bool):
            expression = int(expression)
        if isinstance(expression, Number):
            this_shape = [self.root_dataset.sizes.get(i) for i in self.root_dims]
            result = xr.DataArray(
                np.broadcast_to(expression, this_shape), dims=self.root_dims
            )
            expression = str(expression)
        else:
            if not isinstance(expression, str):
                raise TypeError(
                    f"expression must be a string, not a {type(expression)}"
                )
            if engine is None:
                try:
                    result = self.get_expr(
                        expression,
                        "numexpr",
                        allow_native=False,
                        dtype=dtype,
                        with_coords=with_coords,
                    )
                except Exception:
                    result = self.get_expr(
                        expression,
                        "sharrow",
                        allow_native=False,
                        dtype=dtype,
                        with_coords=with_coords,
                    )
            else:
                result = self.get_expr(
                    expression,
                    engine,
                    allow_native=False,
                    dtype=dtype,
                    with_coords=with_coords,
                )
        if with_coords and "expressions" not in result.coords:
            # add the expression as a scalar coordinate (with no dimension)
            result = result.assign_coords(expressions=xr.DataArray(expression))
        if name is not None:
            result.name = name
        return result

    def eval_many(
        self,
        expressions: Sequence[str] | Mapping[str, str] | pd.Series,
        *,
        engine: Literal[None, "numexpr", "sharrow", "python"] = None,
        dtype=None,
        result_type: Literal["dataset", "dataarray"] = "dataset",
        with_coords: bool = True,
    ) -> xr.Dataset | xr.DataArray:
        """
        Evaluate multiple expressions.

        Parameters
        ----------
        expressions : Sequence[str] or Mapping[str,str] or pd.Series
            The expressions to evaluate.  If a sequence, the names of the
            resulting DataArrays will be the same as the expressions.  If a
            mapping or Series, the keys or index will be used as the names.
        engine : {None, 'numexpr', 'sharrow', 'python'}
            The engine used to resolve expressions. If None, the default is to
            try 'numexpr' first, then 'sharrow' if that fails.
        dtype : str or dtype, optional
            The dtype to use for the result.  If the engine is `sharrow` and
            no value is given, this will default to `float32`, otherwise the
            default is to use the dtype of the result of the concatenation of
            the expressions.
        result_type : {'dataset', 'dataarray'}
            Whether to return a Dataset (with a variable for each expression)
            or a DataArray (with a dimension across all expressions).

        Returns
        -------
        Dataset or DataArray
        """
        if result_type not in {"dataset", "dataarray"}:
            raise ValueError("result_type must be one of ['dataset', 'dataarray']")
        if not isinstance(expressions, (Mapping, pd.Series)):
            expressions = pd.Series(expressions, index=expressions)
        if isinstance(expressions, Mapping):
            expressions = pd.Series(expressions)
        if len(expressions) == 0:
            raise ValueError("no expressions provided")
        if result_type == "dataset":
            arrays = {}
            for k, v in expressions.items():
                a = self.eval(
                    v, engine=engine, dtype=dtype, name=k, with_coords=with_coords
                )
                if "expressions" in a.coords:
                    a = a.drop_vars("expressions")
                arrays[k] = a.assign_attrs(expression=v)
            result = xr.Dataset(arrays)
        else:
            arrays = {}
            for k, v in expressions.items():
                a = self.eval(
                    v, engine=engine, dtype=dtype, name=k, with_coords=with_coords
                )
                if "expressions" in a.coords:
                    a = a.drop_vars("expressions")
                a = a.expand_dims("expressions", -1)
                arrays[k] = a
            result = xr.concat(list(arrays.values()), "expressions")
            if with_coords:
                result = result.assign_coords(
                    expressions=expressions.index,
                    source=xr.DataArray(expressions.values, dims="expressions"),
                )
        return result

    @property
    def subspaces(self):
        """Mapping[str,Dataset] : Direct access to node Dataset objects by name."""
        spaces = {}
        for k in self._graph.nodes:
            s = self._graph.nodes[k].get("dataset", None)
            if s is not None:
                spaces[k] = s
        return spaces

    def subspaces_iter(self):
        for k in self._graph.nodes:
            s = self._graph.nodes[k].get("dataset", None)
            if s is not None:
                yield (k, s)

    def contains_subspace(self, key) -> bool:
        """
        Is this named Dataset in this tree's subspaces.

        Parameters
        ----------
        key : str

        Returns
        -------
        bool
        """
        return key in self._graph.nodes

    def get_subspace(self, key, default_empty=False) -> xr.Dataset:
        """
        Access named Dataset from this tree's subspaces.

        Parameters
        ----------
        key : str
        default_empty : bool, default False
            Return an empty Dataset if the key is not found.

        Returns
        -------
        xr.Dataset
        """
        result = self._graph.nodes[key].get("dataset", None)
        if result is None:
            if default_empty:
                result = xr.Dataset()
            else:
                raise KeyError(key)
        return result

    def namespace_names(self):
        namespace = set()
        for spacename, spacearrays in self.subspaces_iter():
            for k, _arr in spacearrays.coords.items():
                namespace.add(f"__{spacename or 'base'}__{k}")
            for k, _arr in spacearrays.items():
                if k.startswith("_s_"):
                    namespace.add(f"__{spacename or 'base'}__{k}__indptr")
                    namespace.add(f"__{spacename or 'base'}__{k}__indices")
                    namespace.add(f"__{spacename or 'base'}__{k}__data")
                else:
                    namespace.add(f"__{spacename or 'base'}__{k}")
        return namespace

    @property
    def dims(self):
        """Mapping from dimension names to lengths across all dataset nodes."""
        dims = {}
        for _k, v in self.subspaces_iter():
            for name, length in v.sizes.items():
                if name in dims:
                    if dims[name] != length:
                        raise ValueError(
                            "inconsistent dimensions\n" + self.dims_detail()
                        )
                else:
                    dims[name] = length
        return xr.core.utils.Frozen(dims)

    sizes = dims  # alternate name

    def dims_detail(self):
        """
        Report on the names and sizes of dimensions in all Dataset nodes.

        Returns
        -------
        str
        """
        s = ""
        for k, v in self.subspaces_iter():
            s += f"\n{k}:"
            for name, length in v.sizes.items():
                s += f"\n - {name}: {length}"
        return s[1:]

    def drop_dims(self, dims, inplace=False, ignore_missing_dims=True):
        """
        Drop dimensions from root Dataset node.

        Parameters
        ----------
        dims : str or Iterable[str]
            One or more named dimensions to drop.
        inplace : bool, default False
            Whether to drop dimensions in-place.
        ignore_missing_dims : bool, default True
            Simply ignore any dimensions that are not present.

        Returns
        -------
        DataTree
            Returns self if dropping inplace, otherwise returns a copy
            with dimensions dropped.
        """
        if isinstance(dims, str):
            dims = [dims]
        if inplace:
            obj = self
        else:
            obj = self.copy()

        if not ignore_missing_dims:
            new_root_dataset = obj.root_dataset.drop_dims(dims)
        else:
            new_root_dataset = obj.root_dataset
            for d in dims:
                if d in obj.root_dataset.dims:
                    new_root_dataset = new_root_dataset.drop_dims(d)

        # remove subspaces that rely on dropped dim
        boot_queue = set()
        booted = set()
        for (up, dn, _n), e in obj._graph.edges.items():
            if up == obj.root_node_name:
                _analog = e.get("analog", "<missing>")
                if _analog in dims:
                    boot_queue.add(dn)
                if _analog != "<missing>" and _analog not in new_root_dataset:
                    boot_queue.add(dn)
                if e.get("parent_name", "<missing>") in dims:
                    boot_queue.add(dn)
                if e.get("parent_name", "<missing>") not in new_root_dataset:
                    boot_queue.add(dn)

        while boot_queue:
            b = boot_queue.pop()
            booted.add(b)
            for up, dn, _n in obj._graph.edges.keys():
                if up == b:
                    boot_queue.add(dn)

        edges_to_remove = [e for e in obj._graph.edges if e[1] in booted]
        obj._graph.remove_edges_from(edges_to_remove)
        obj._graph.remove_nodes_from(booted)

        obj.root_dataset = new_root_dataset
        obj.dim_order = tuple(x for x in self.dim_order if x not in dims)
        return obj

    def get_indexes(
        self,
        position_only=True,
        as_dict=True,
        replacements=None,
        use_cache=True,
        check_shapes=True,
    ):
        if use_cache and (position_only, as_dict) in self._cached_indexes:
            return self._cached_indexes[(position_only, as_dict)]
        if not position_only:
            raise NotImplementedError
        dims = [
            d
            for d in self.dims
            if d[-1:] != "_" or (d[-1:] == "_" and d[:-1] not in self.dims)
        ]
        if replacements is not None:
            obj = self.replace_datasets(replacements)
        else:
            obj = self
        result = {}
        result_shape = None
        for k in sorted(dims):
            result_k = obj._getitem(k, include_blank_dims=True, only_dims=True)
            if result_shape is None:
                result_shape = result_k.shape
            if result_shape != result_k.shape:
                if check_shapes:
                    raise ValueError(
                        f"inconsistent index shapes {result_k.shape} v {result_shape} "
                        f"(probably an error on {k} or {sorted(dims)[0]})"
                    )
            result[k] = result_k

        if as_dict:
            result = {k: _dataarray_to_numpy(v) for k, v in result.items()}
        else:
            result = Dataset(result)
        if use_cache:
            self._cached_indexes[(position_only, as_dict)] = result
        return result

    def replace_datasets(self, other=None, validate=True, redigitize=True, **kwargs):
        """
        Replace one or more datasets in the nodes of this tree.

        Parameters
        ----------
        other : Mapping[str,Dataset]
            A dictionary of replacement datasets.
        validate : bool, default True
            Raise an error when replacing downstream datasets that
            are referenced by position, unless the replacement is identically
            sized.  If validation is deactivated, and an incompatible dataset
            is placed in this tree, flows that rely on that relationship will
            give erroneous results or crash with a segfault.
        redigitize : bool, default True
            Automatically re-digitize relationships that are label-based and
            were previously digitized.
        **kwargs : Mapping[str,Dataset]
            Alternative format to `other`.

        Returns
        -------
        DataTree
            A new DataTree with data replacements completed.
        """
        replacements = {}
        if other is not None:
            replacements.update(other)
        replacements.update(kwargs)
        graph = self._graph.copy()
        for k in replacements:
            if k not in graph.nodes:
                raise KeyError(k)
            x = construct(replacements[k])
            if validate:
                if x.sizes != graph.nodes[k]["dataset"].sizes:
                    # when replacement dimensions do not match, check for
                    # any upstream nodes that reference this dataset by
                    # position... which will potentially be problematic.
                    for e in self._graph.edges:
                        if e[1] == k:
                            indexing = self._graph.edges[e].get("indexing")
                            if indexing == "position":
                                raise ValueError(
                                    f"dimensions mismatch on "
                                    f"positionally-referenced dataset {k}: "
                                    f"receiving {x.dims} "
                                    f"expected {graph.nodes[k]['dataset'].dims}"
                                )
                # also if any dim coordinates are changing, redigitize
                for dim in x.dims:
                    if dim in graph.nodes[k]["dataset"].coords:
                        if not np.array_equal(
                            graph.nodes[k]["dataset"].coords[dim].data,
                            x.coords[dim].data,
                        ):
                            # find all edges with digitized label relationships
                            # and cast them back to label
                            for e in graph.edges:
                                if e[1] == k:
                                    r = self._get_relationship(e)
                                    if r.child_name == dim and r.analog:
                                        graph.edges[e]["indexing"] = "label"
                                        graph.edges[e]["parent_name"] = r.analog
            if k in self.replacement_filters:
                x = self.replacement_filters[k](x)
            graph.nodes[k]["dataset"] = x
        result = type(self)(graph, self.root_node_name, **self.__shallow_copy_extras())
        if redigitize:
            result.digitize_relationships(inplace=True)
        return result

    def setup_flow(
        self,
        definition_spec,
        *,
        cache_dir=None,
        name=None,
        dtype="float32",
        boundscheck=False,
        error_model="numpy",
        nopython=True,
        fastmath=True,
        parallel=True,
        readme=None,
        flow_library=None,
        extra_hash_data=(),
        write_hash_audit=True,
        hashing_level=1,
        dim_exclude=None,
        with_root_node_name=None,
    ):
        """
        Set up a new Flow for analysis using the structure of this DataTree.

        Parameters
        ----------
        definition_spec : Dict[str,str]
            Gives the names and expressions that define the variables to
            create in this new `Flow`.
        cache_dir : Path-like, optional
            A location to write out generated python and numba code. If not
            provided, a unique temporary directory is created.
        name : str, optional
            The name of this Flow used for writing out cached files. If not
            provided, a unique name is generated. If `cache_dir` is given,
            be sure to avoid name conflicts with other flow's in the same
            directory.
        dtype : str, default "float32"
            The name of the numpy dtype that will be used for the output.
        boundscheck : bool, default False
            If True, boundscheck enables bounds checking for array indices, and
            out of bounds accesses will raise IndexError. The default is to not
            do bounds checking, which is faster but can produce garbage results
            or segfaults if there are problems, so try turning this on for
            debugging if you are getting unexplained errors or crashes.
        error_model : {'numpy', 'python'}, default 'numpy'
            The error_model option controls the divide-by-zero behavior. Setting
            it to ‘python’ causes divide-by-zero to raise exception like
            CPython. Setting it to ‘numpy’ causes divide-by-zero to set the
            result to +/-inf or nan.
        nopython : bool, default True
            Compile using numba's `nopython` mode.  Provided for debugging only,
            as there's little point in turning this off for production code, as
            all the speed benefits of sharrow will be lost.
        fastmath : bool, default True
            If true, fastmath enables the use of "fast" floating point transforms,
            which can improve performance but can result in tiny distortions in
            results.  See numba docs for details.
        parallel : bool, default True
            Enable or disable parallel computation for certain functions.
        readme : str, optional
            A string to inject as a comment at the top of the flow Python file.
        flow_library : Mapping[str,Flow], optional
            An in-memory cache of precompiled Flow objects.  Using this can result
            in performance improvements when repeatedly using the same definitions.
        extra_hash_data : Tuple[Hashable], optional
            Additional data used for generating the flow hash.  Useful to prevent
            conflicts when using a flow_library with multiple similar flows.
        write_hash_audit : bool, default True
            Writes a hash audit log into a comment in the flow Python file, for
            debugging purposes.
        hashing_level : int, default 1
            Level of detail to write into flow hashes.  Increase detail to avoid
            hash conflicts for similar flows.  Level 2 adds information about
            names used in expressions and digital encodings to the flow hash,
            which prevents conflicts but requires more pre-computation to generate
            the hash.
        dim_exclude : Collection[str], optional
            Exclude these root dataset dimensions from this flow.

        Returns
        -------
        Flow
        """
        from .flows import Flow

        return Flow(
            self,
            definition_spec,
            cache_dir=cache_dir or self.cache_dir,
            name=name,
            dtype=dtype,
            boundscheck=boundscheck,
            nopython=nopython,
            fastmath=fastmath,
            parallel=parallel,
            readme=readme,
            flow_library=flow_library,
            extra_hash_data=extra_hash_data,
            hashing_level=hashing_level,
            error_model=error_model,
            write_hash_audit=write_hash_audit,
            dim_order=self.dim_order,
            dim_exclude=dim_exclude,
            with_root_node_name=with_root_node_name,
        )

    def get_named_array(self, mangled_name):
        if mangled_name[:2] != "__":
            raise KeyError(mangled_name)
        name1, name2 = mangled_name[2:].split("__", 1)
        if name1 == "aux_var":
            return self.aux_vars[name2]
        dataset = self._graph.nodes[name1].get("dataset")
        if name2.startswith("_s_"):
            if name2.endswith("__data"):
                return dataset[name2[:-6]].data.data
            elif name2.endswith("__indptr"):
                return dataset[name2[:-8]].data.indptr
            elif name2.endswith("__indices"):
                return dataset[name2[:-9]].data.indices
        try:
            _d = dataset[name2]
        except KeyError as err:
            raise KeyError(f"{name1}.{name2}") from err
        else:
            return _dataarray_to_numpy(_d)

    _BY_OFFSET = "digitizedOffset"

    def digitize_relationships(self, inplace=False, redigitize=True):
        """
        Convert all label-based relationships into position-based.

        Parameters
        ----------
        inplace : bool, default False
        redigitize : bool, default True
            Re-compute position-based relationships from labels, even
            if the relationship had previously been digitized.

        Returns
        -------
        DataTree or None
            Only returns a copy if not digitizing in-place.
        """
        if inplace:
            obj = self
        else:
            obj = self.copy()

        for e in obj._graph.edges:
            r = obj._get_relationship(e)
            if redigitize and r.analog:
                p_dataset = obj._graph.nodes[r.parent_data].get("dataset", None)
                if p_dataset is not None:
                    if r.parent_name not in p_dataset:
                        r.indexing = "label"
                        r.parent_name = r.analog
            if r.indexing == "label":
                p_dataset = obj._graph.nodes[r.parent_data].get("dataset", None)
                if p_dataset is None:
                    raise ValueError(f"no dataset found for {r.parent_data}")
                c_dataset = obj._graph.nodes[r.child_data].get("dataset", None)
                if c_dataset is None:
                    raise ValueError(f"no dataset found for {r.child_data}")

                upstream = p_dataset[r.parent_name]
                downstream = c_dataset[r.child_name]

                upstream_is_categorical = (
                    isinstance(upstream, xr.DataArray) and upstream.cat.is_categorical()
                )
                # check if both upstream and downstream are categoricals with the same categories
                if upstream_is_categorical:
                    if np.array_equal(upstream.cat.category_array(), downstream):
                        # if so, we can just use the codes
                        offsets = upstream
                        if (offsets < 0).any():
                            raise ValueError(
                                f"detected missing values in digitizing {r.parent_data}.{r.parent_name}"
                            )
                    else:
                        raise ValueError(
                            f"upstream ({r.parent_data}.{r.parent_name}) and "
                            f"downstream ({r.child_data}.{r.child_name}) categoricals "
                            f"have different categories"
                        )
                else:
                    # vectorize version
                    mapper = {
                        i: j for (j, i) in enumerate(_dataarray_to_numpy(downstream))
                    }

                    def mapper_get(x, mapper=mapper):
                        return mapper.get(x, 0)

                    if upstream.size:
                        offsets = xr.apply_ufunc(np.vectorize(mapper_get), upstream)
                    else:
                        offsets = xr.DataArray([], dims=["index"])
                    if offsets.dtype.kind != "i":
                        warnings.warn(
                            f"detected missing values in digitizing {r.parent_data}.{r.parent_name}",
                            stacklevel=2,
                        )

                # candidate name for write back
                r_parent_name_new = (
                    f"{self._BY_OFFSET}{r.parent_name}_{r.child_data}_{r.child_name}"
                )

                # it is common to have mirrored offsets in various dimensions.
                # we'd like to retain only the same data in memory once, so we'll
                # check if these offsets match any existing ones and if so just
                # point to that memory.
                for k in p_dataset:
                    if isinstance(k, str) and k.startswith(self._BY_OFFSET):
                        if p_dataset[k].equals(offsets):
                            # we found a match, so we'll assign this name to
                            # the match's memory storage instead of replicating it.
                            obj._graph.nodes[r.parent_data]["dataset"] = (
                                p_dataset.assign({r_parent_name_new: p_dataset[k]})
                            )
                            # r_parent_name_new = k
                            break
                else:
                    # no existing offset arrays match, make this new one
                    obj._graph.nodes[r.parent_data]["dataset"] = p_dataset.assign(
                        {r_parent_name_new: offsets}
                    )
                obj._graph.edges[e].update(
                    dict(
                        parent_name=r_parent_name_new,
                        indexing="position",
                        analog=r.parent_name,
                    )
                )

        if not inplace:
            return obj

    @property
    def relationships_are_digitized(self):
        """Bool : Whether all relationships are digital (by position)."""
        for e in self._graph.edges:
            r = self._get_relationship(e)
            if r.indexing != "position":
                return False
        return True

    def _arg_tokenizer(
        self, spacename, spacearray, spacearrayname, exclude_dims=None, blends=None
    ):
        if blends is None:
            blends = {}

        if spacename == self.root_node_name:
            root_dataset = self.root_dataset
            from .flows import presorted

            root_dims = list(presorted(root_dataset.dims, self.dim_order, exclude_dims))
            if isinstance(spacearray, str):
                from_dims = root_dataset[spacearray].dims
            else:
                from_dims = spacearray.dims
            return (
                tuple(
                    ast.parse(f"_arg{root_dims.index(dim):02}", mode="eval").body
                    for dim in from_dims
                ),
                blends,
            )

        if isinstance(spacearray, str):
            spacearray_ = self._graph.nodes[spacename]["dataset"][spacearray]
        else:
            spacearray_ = spacearray

        from_dims = spacearray_.dims
        offset_source = spacearray_.attrs.get("digital_encoding", {}).get(
            "offset_source", None
        )
        if offset_source is not None:
            from_dims = self._graph.nodes[spacename]["dataset"][offset_source].dims

        tokens = []

        n_missing_tokens = 0
        for dimname in from_dims:
            found_token = False
            for e in self._graph.in_edges(spacename, keys=True):
                this_dim_name = self._graph.edges[e]["child_name"]
                retarget = None
                if dimname != this_dim_name:
                    retarget = self._graph.nodes[spacename][
                        "dataset"
                    ].redirection.target(this_dim_name)
                    if dimname != retarget:
                        continue
                parent_name = self._graph.edges[e]["parent_name"]
                parent_data = e[0]

                upside_ast, blends_ = self._arg_tokenizer(
                    parent_data,
                    parent_name,
                    spacearrayname=spacearrayname,
                    exclude_dims=exclude_dims,
                    blends=blends,
                )
                try:
                    upside = ", ".join(unparse(t) for t in upside_ast)
                except:  # noqa: E722
                    if self.root_node_name is False:
                        upside = None
                    else:
                        print(f"{parent_data=}")
                        print(f"{parent_name=}")
                        print(f"{spacearrayname=}")
                        print(f"{exclude_dims=}")
                        print(f"{blends=}")
                        for t in upside_ast:
                            str_t = str(t)
                            if len(str_t) < 2000:
                                print(f"t:{str_t}")
                            else:
                                print(f"t:{str_t[:200]}...")
                        raise
                if upside is not None:
                    # check for redirection target
                    if retarget is not None:
                        tokens.append(
                            f"__{spacename}___digitized_{retarget}_of_{this_dim_name}[__{parent_data}__{parent_name}[{upside}]]"
                        )
                    else:
                        tokens.append(f"__{parent_data}__{parent_name}[{upside}]")
                found_token = True
                break
            if not found_token:
                if dimname in self.subspaces[spacename].indexes:
                    if self.root_node_name is False:
                        tokens.append(False)
                    else:
                        ix = self.subspaces[spacename].indexes[dimname]
                        ix = {i: n for n, i in enumerate(ix)}
                        tokens.append(ix)
                        n_missing_tokens += 1
                elif dimname.endswith("_indices") or dimname.endswith("_indptr"):
                    tokens.append(None)
                    # this dimension corresponds to a blender

        if n_missing_tokens > 1:
            raise ValueError("at most one missing dimension is allowed")
        result = []
        for t in tokens:
            if isinstance(t, str):
                # print(f"TOKENIZE: {spacename=} {spacearray=} {t}")
                result.append(ast.parse(t, mode="eval").body)
            else:
                result.append(t)
        return tuple(result), blends

    @property
    def coords(self):
        return self.root_dataset.coords

    def get_index(self, dim):
        for _spacename, subspace in self.subspaces.items():
            if dim in subspace.coords:
                return subspace.indexes[dim]

    def copy(self):
        return type(self)(
            self._graph.copy(), self.root_node_name, **self.__shallow_copy_extras()
        )

    def all_var_names(self, uniquify=False, _duplicated_names=None):
        ordered_names = []
        require_unique = _duplicated_names is None and not uniquify
        need_second_pass = _duplicated_names is None and uniquify
        print(f"{require_unique=}")
        discovered_names = set()
        duplicated_names = _duplicated_names or set()
        for spacename, space in self.subspaces_iter():
            for name in space.variables:
                if name in duplicated_names:
                    if require_unique:
                        raise ValueError(f"duplicate name {name}")
                    elif uniquify:
                        ordered_names.append(f"{spacename}.{name}")
                    else:
                        ordered_names.append(name)
                elif name in discovered_names:
                    duplicated_names.add(name)
                    if require_unique:
                        raise ValueError(f"duplicate name {name}")
                    else:
                        ordered_names.append(name)
                else:
                    discovered_names.add(name)
                    ordered_names.append(name)
        if need_second_pass:
            return self.all_var_names(uniquify=True, _duplicated_names=duplicated_names)
        return ordered_names

    def merged_dataset(self, columns=None, uniquify=False):
        if columns is None:
            columns = self.all_var_names(uniquify=uniquify)
        if len(self.root_dataset.dims) > 1:
            raise NotImplementedError("only single dim root datasets")
        dim_name = self.root_dataset.single_dim.dim_name
        vx = []
        coords = {}
        for k in columns:
            v = self._getitem(k).single_dim.rename(dim_name)
            if v.name == v.dims[0]:
                coords[v.name] = v
            else:
                vx.append(v)
        result = xr.merge(vx, compat="override", join="override")
        if coords:
            result.assign_coords(coords)
        return result

    def __iter__(self):
        """Iterate over all the datasets."""
        import itertools

        if hasattr(self, "_eval_cache"):
            z = (self._eval_cache,)
        else:
            z = ()
        return itertools.chain(*z, *(v for k, v in self.subspaces_iter()))

    def __setitem__(self, key, value):
        if hasattr(self, "_eval_cache"):
            self._eval_cache[key] = value
        else:
            raise NotImplementedError("setitem not supported")



def example(extract='m', estimate=False):
    import larch

    larch.__version__

    d = larch.examples.MTC(format="dataset")
    d

    m = larch.Model(d, compute_engine="numba")

    from larch import P, X

    for a in [2, 3]:
        m.utility_co[a] = +P("hhinc#2,3") * X("hhinc")
    for a in [4, 5, 6]:
        m.utility_co[a] = +P(f"hhinc#{a}") * X("hhinc")

    d.dc.alts_mapping

    for a, name in d.dc.alts_mapping.items():
        if a == 1:
            continue
        m.utility_co[a] += (
            +P("ASC_" + name)
            + P("vehbywrk_" + name) * X("vehbywrk")
            + P("wkcbd_" + name) * X("wkccbd + wknccbd")
            + P("wkempden_" + name) * X("wkempden")
        )

    m.utility_ca = (
        +P("nonmotorized_time") * X("(altid> 4) * tottime")
        + P("motorized_time") * X("(altid <= 4) * ivtt")
        + (P("motorized_time") + (P("motorized_ovtbydist") / X("dist")))
        * X("(altid <= 4) * ovtt")
        + P("costbyinc") * X("totcost/hhinc")
    )

    m.availability_ca_var = "avail"
    m.choice_ca_var = "chose"

    m.title = "MTC Example 16, Cost by Income"

    m.choice_avail_summary()

    m.set_cap(20)

    assert m.compute_engine == "numba"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(stderr=True, method="bhhh")
    m.calculate_parameter_covariance()
    m.loglike()

    m.parameter_summary()

    m.ordering = (
        (
            "LOS",
            ".*cost.*",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Income",
            "hhinc.*",
        ),
        (
            "Ownership",
            "vehbywrk.*",
        ),
        (
            "Zonal",
            "wkcbd.*",
            "wkempden.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.parameter_summary()

    m.estimation_statistics()

    # TEST
    revealed_x = dict(zip(m.pnames, result.x))

    # TEST
    from pytest import approx

    expected_x = {
        "ASC_Bike": -1.6217774612007492,
        "ASC_SR2": -1.7297986582095874,
        "ASC_SR3+": -3.6562561506975255,
        "ASC_Transit": -0.6917042404795228,
        "ASC_Walk": 0.07521549340170937,
        "costbyinc": -0.05177365897066612,
        "hhinc#2,3": 3.691866127922073e-05,
        "hhinc#4": -0.005335573957018,
        "hhinc#5": -0.008671987053899464,
        "hhinc#6": -0.006017166395409271,
        "motorized_ovtbydist": -0.13272166895031992,
        "motorized_time": -0.0201578217327816,
        "nonmotorized_time": -0.045438649548417406,
        "vehbywrk_Bike": -0.7040646431527947,
        "vehbywrk_SR2": -0.38161738059345535,
        "vehbywrk_SR3+": -0.13880487622153995,
        "vehbywrk_Transit": -0.937505306213354,
        "vehbywrk_Walk": -0.72385335687274,
        "wkcbd_Bike": 0.4863239419169535,
        "wkcbd_SR2": 0.24714212608546096,
        "wkcbd_SR3+": 1.0943587942374322,
        "wkcbd_Transit": 1.3056155684640574,
        "wkcbd_Walk": 0.09724802958992065,
        "wkempden_Bike": 0.0019224683185795574,
        "wkempden_SR2": 0.0015964358018843575,
        "wkempden_SR3+": 0.0022037880176673103,
        "wkempden_Transit": 0.0031316557035939658,
        "wkempden_Walk": 0.002881429335761498,
    }
    for k in expected_x:
        assert revealed_x[k] == approx(
            expected_x[k], 2e-2
        ), f"{k}, {revealed_x[k]/expected_x[k]}"
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



# based on https://gist.github.com/scivision/ad241e9cf0474e267240e196d7545eca

import os
import sys
import tarfile
import tempfile
from pathlib import Path

try:
    import zstandard  # pip install zstandard
except ModuleNotFoundError:
    zstandard = None


def extract_zst(archive: Path, out_path: Path):
    """
    Extract content of zst file to a target file system directory.

    works on Windows, Linux, MacOS, etc.

    Parameters
    ----------
    archive: pathlib.Path or str
      .zst file to extract
    out_path: pathlib.Path or str
      directory to extract files and directories to
    """
    if zstandard is None:
        raise ImportError("pip install zstandard")

    archive = Path(archive).expanduser()
    out_path = Path(out_path).expanduser().resolve()
    # need .resolve() in case intermediate relative dir doesn't exist

    dctx = zstandard.ZstdDecompressor()

    with tempfile.TemporaryFile(suffix=".tar") as ofh:
        with archive.open("rb") as ifh:
            dctx.copy_stream(ifh, ofh)
        ofh.seek(0)
        with tarfile.open(fileobj=ofh) as z:
            z.extractall(out_path)


def compress_zst(in_path: Path, archive: Path):
    """
    Compress a directory into a .tar.zst file.

    Certain hidden files are excluded, including .git directories and
    macOS's .DS_Store files.

    Parameters
    ----------
    in_path: pathlib.Path or str
      directory to compress
    archive: pathlib.Path or str
      .tar.zst file to compress into
    """
    if zstandard is None:
        raise ImportError("pip install zstandard")
    dctx = zstandard.ZstdCompressor(level=9, threads=-1, write_checksum=True)
    with tempfile.TemporaryFile(suffix=".tar") as ofh:
        with tarfile.open(fileobj=ofh, mode="w") as z:
            for dirpath, dirnames, filenames in os.walk(in_path):
                if os.path.basename(dirpath) == ".git":
                    continue
                for n in range(len(dirnames) - 1, -1, -1):
                    if dirnames[n] == ".git" or dirnames[n].startswith("---"):
                        dirnames.pop(n)
                for f in filenames:
                    if f.startswith(".git") or f == ".DS_Store" or f.startswith("---"):
                        continue
                    finame = Path(os.path.join(dirpath, f))
                    arcname = finame.relative_to(in_path)
                    print(f"> {arcname}")
                    z.add(finame, arcname=arcname)
        ofh.seek(0)
        with archive.open("wb") as ifh:
            dctx.copy_stream(ofh, ifh)


if __name__ == "__main__":
    x = Path(sys.argv[1])
    name = x.name
    if name.endswith(".tar.zst"):
        y = x.with_name(name[:-8])
        if x.exists():
            if not y.exists():
                print(f"extracting from: {x}")
                extract_zst(x, y)
            else:
                print(f"not extracting, existing target: {y}")
        else:
            print(f"not extracting, does not exist: {x}")
    else:
        y = x.with_name(name + ".tar.zst")
        if x.exists():
            if not y.exists():
                print(f"compressing to tar.zst: {x}")
                compress_zst(x, y)
            else:
                print(f"not compressing, existing tar.zst: {x}")
        else:
            print(f"not compressing, does not exist: {x}")



from __future__ import annotations

import datetime
import logging
import sys
import time
from contextlib import contextmanager

LOGGER_NAME = __name__.split(".")[0]
FILE_LOG_FORMAT = "%(name)s.%(levelname)s: %(message)s"
CONSOLE_LOG_FORMAT = "{asctime} [{elapsedTime}] {name:s}.{levelname:s}: {message:s}"
DEFAULT_LOG_LEVEL = logging.INFO


def timesize_single(t):
    if t < 60:
        return f"{t:.2f}s"
    elif t < 3600:
        return f"{t / 60:.2f}m"
    elif t < 86400:
        return f"{t / 3600:.2f}h"
    else:
        return f"{t / 86400:.2f}d"


def timesize_stack(t):
    if t < 60:
        return f"{t:.2f}s"
    elif t < 3600:
        return f"{t // 60:.0f}m {timesize_stack(t % 60)}"
    elif t < 86400:
        return f"{t // 3600:.0f}h {timesize_stack(t % 3600)}"
    else:
        return f"{t // 86400:.0f}d {timesize_stack(t % 86400)}"


class ElapsedTimeFormatter(logging.Formatter):
    def format(self, record):
        duration_milliseconds = record.relativeCreated
        hours, rem = divmod(duration_milliseconds / 1000, 3600)
        minutes, seconds = divmod(rem, 60)
        if hours:
            record.elapsedTime = f"{int(hours):0>2}:{int(minutes):0>2}:{seconds:07.4f}"
        else:
            record.elapsedTime = f"{int(minutes):0>2}:{seconds:07.4f}"
        return super().format(record)


def log_to_console(level=None):
    if level is None:
        level = DEFAULT_LOG_LEVEL

    logger = logging.getLogger(LOGGER_NAME)

    if level < logger.level or logger.level == logging.NOTSET:
        logger.setLevel(level)

    # avoid creation of multiple stream handlers for logging to console
    for entry in logger.handlers:
        if (isinstance(entry, logging.StreamHandler)) and (
            entry.formatter._fmt == CONSOLE_LOG_FORMAT
        ):
            if level < entry.level:
                entry.setLevel(level)
            logger.propagate = False
            return logger

    console_handler = logging.StreamHandler(stream=sys.stderr)
    console_handler.setLevel(level)
    formatter = ElapsedTimeFormatter(CONSOLE_LOG_FORMAT, style="{")
    formatter.default_time_format = "%H:%M:%S"
    formatter.default_msec_format = "%s.%03d"
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    try:
        logging_start_time = datetime.datetime.fromtimestamp(
            logging._startTime
        ).strftime("%Y-%m-%d %I:%M:%S %p")
    except Exception:
        pass
    else:
        logger.info(f"Logging started {logging_start_time}")

    return logger


def log_to_file(filename, level=None):
    if level is None:
        level = DEFAULT_LOG_LEVEL

    logger = logging.getLogger(LOGGER_NAME)

    # avoid creation of multiple file handlers for logging to the same file
    for entry in logger.handlers:
        if (isinstance(entry, logging.FileHandler)) and (
            entry.baseFilename == filename
        ):
            return logger

    file_handler = logging.FileHandler(filename)
    file_handler.setLevel(level)
    file_handler.setFormatter(logging.Formatter(FILE_LOG_FORMAT))
    logger.addHandler(file_handler)

    return logger


logger = log = log_to_console()


@contextmanager
def timing_log(label=""):
    start_time = time.time()
    log.critical(f"<TIME BEGINS> {label}")
    try:
        yield
    except Exception:  # noqa: E722
        log.critical(
            f"<TIME ERROR!> {label} <{timesize_stack(time.time() - start_time)}>"
        )
        raise
    else:
        log.critical(
            f"< TIME ENDS > {label} <{timesize_stack(time.time() - start_time)}>"
        )


class TimingLog:
    def __init__(self, label="", log=None, level=50):
        global logger
        if log is None:
            log = logger
        self.label = label
        self.log = log
        self.level = level
        self.split_time = None
        self.current_task = ""

    def __enter__(self):
        self.start_time = time.time()
        self.log.log(self.level, f"<BEGIN> {self.label}")
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        now = time.time()
        if self.split_time is not None:
            self.log.log(
                self.level,
                f"<SPLIT> {self.label} / Final <{timesize_stack(now - self.split_time)}>",
            )
        if exc_type is None:
            self.log.log(
                self.level,
                f"<-END-> {self.label} <{timesize_stack(now - self.start_time)}>",
            )
        else:
            self.log.log(
                self.level,
                f"<ERROR> {self.label} <{timesize_stack(now - self.start_time)}>",
            )

    def split(self, note=""):
        if self.split_time is None:
            self.split_time = self.start_time
        now = time.time()
        if note:
            note = " / " + note
        self.log.log(
            self.level,
            f"<SPLIT> {self.label}{note} <{timesize_stack(now - self.split_time)}>",
        )
        self.split_time = now



from __future__ import annotations

import numpy as np
import pandas as pd
from pandas.api.types import is_object_dtype, is_string_dtype


def clean_summary(df, root_id=None):
    if root_id is not None:
        totals = df.loc[root_id, :]
        df.drop(index=root_id, inplace=True)
    else:
        totals = df.sum()

    relevant_totals = (
        "chosen",
        "chosen weighted",
        "chosen unweighted",
        "chosen but not available",
        "chosen but not available weighted",
        "chosen but not available unweighted",
        "chosen thus available",
        "not available so not chosen",
    )

    for tot in relevant_totals:
        if tot in totals:
            df.loc["< Total All Alternatives >", tot] = totals[tot]

    for i in df.columns:
        if i not in relevant_totals:
            if is_object_dtype(df[i].dtype) or is_string_dtype(df[i].dtype):
                df.loc["< Total All Alternatives >", i] = ""
    df.drop("_root_", errors="ignore", inplace=True)

    if "availability condition" in df:
        df["availability condition"] = df["availability condition"].fillna("")

    for i in (
        "chosen",
        "chosen but not available",
        "chosen thus available",
        "not available so not chosen",
    ):
        if i in df.columns and all(df[i] == df[i].astype(int)):
            df[i] = df[i].astype(int)

    for i in ("available",):
        if i in df.columns:
            df[i] = df[i].astype(pd.Int64Dtype())

    return df


def choice_avail_summary(
    dataset, graph=None, availability_co_vars=None, streaming=None
):
    """
    Generate a summary of choice and availability statistics.

    Parameters
    ----------
    dataset : Dataset
        The loaded dataset to summarize, which should have
        `ch` and `av` variables.
    graph : NestingTree, optional
        The nesting graph.
    availability_co_vars : dict, optional
        Also attach the definition of the availability conditions.

    Returns
    -------
    pandas.DataFrame
    """
    if graph is None:
        if "ch" in dataset:
            ch_ = np.asarray(dataset["ch"].copy())
        else:
            ch_ = None
        av_ = np.asarray(dataset.get("av"))
    else:
        from ..model.cascading import array_av_cascade, array_ch_cascade

        ch_ = array_ch_cascade(dataset.get("ch"), graph)
        av_ = array_av_cascade(dataset.get("av"), graph)
        av_[av_ > 1] = 1

    if ch_ is not None:
        ch = ch_.sum(0)
    else:
        ch = None

    if av_ is not None:
        av = av_.sum(0)
    else:
        av = None

    arr_wt = dataset.get("wt")
    if arr_wt is not None:
        if ch_ is not None:
            ch_w = pd.Series((ch_ * arr_wt.values.reshape(-1, 1)).sum(0))
        else:
            ch_w = None
        if av_ is not None:
            av_w = pd.Series((av_ * arr_wt.values.reshape(-1, 1)).sum(0))
        else:
            av_w = None
        show_wt = np.any(ch != ch_w)
    else:
        ch_w = ch
        av_w = av
        show_wt = False

    if av_ is not None:
        ch_[av_ > 0] = 0
    if ch_ is not None and ch_.sum() > 0:
        ch_but_not_av = ch_.sum(0)
        if arr_wt is not None:
            ch_but_not_av_w = pd.Series((ch_ * arr_wt.values).sum(0), index=ch_.columns)
        else:
            ch_but_not_av_w = ch_but_not_av
    else:
        ch_but_not_av = None
        ch_but_not_av_w = None

    from collections import OrderedDict

    od = OrderedDict()

    if graph is not None:
        idx = graph.standard_sort
        od["name"] = pd.Series(graph.standard_sort_names, index=idx)
    else:
        idx = dataset.dc.altids()
        if dataset.get("alt_names") is not None:
            od["name"] = pd.Series(dataset.get("alt_names"), index=idx)
        elif dataset.get("altnames") is not None:
            od["name"] = pd.Series(dataset.get("altnames"), index=idx)

    if show_wt:
        od["chosen weighted"] = pd.Series(ch_w, index=idx)
        od["chosen unweighted"] = pd.Series(ch, index=idx)
        od["available weighted"] = pd.Series(av_w, index=idx)
        od["available unweighted"] = pd.Series(av, index=idx)
    else:
        od["chosen"] = pd.Series(ch, index=idx)
        od["available"] = pd.Series(av, index=idx)
    if ch_but_not_av is not None:
        if show_wt:
            od["chosen but not available weighted"] = pd.Series(
                ch_but_not_av_w, index=idx
            )
            od["chosen but not available unweighted"] = pd.Series(
                ch_but_not_av, index=idx
            )
        else:
            od["chosen but not available"] = pd.Series(ch_but_not_av, index=idx)

    if availability_co_vars is not None:
        od["availability condition"] = pd.Series(
            availability_co_vars.values(),
            index=availability_co_vars.keys(),
            dtype=np.str_,
        )

    result = pd.DataFrame.from_dict(od)
    result = clean_summary(result, root_id=graph.root_id if graph is not None else None)
    # if graph is not None:
    #     totals = result.loc[graph.root_id, :]
    #     result.drop(index=graph.root_id, inplace=True)
    # else:
    #     totals = result.sum()
    #
    # for tot in (
    #     "chosen",
    #     "chosen weighted",
    #     "chosen unweighted",
    #     "chosen but not available",
    #     "chosen but not available weighted",
    #     "chosen but not available unweighted",
    #     "chosen thus available",
    #     "not available so not chosen",
    # ):
    #     if tot in totals:
    #         result.loc["< Total All Alternatives >", tot] = totals[tot]
    #
    # result.loc[
    #     "< Total All Alternatives >",
    #     pd.isnull(result.loc["< Total All Alternatives >", :]),
    # ] = ""
    # result.drop("_root_", errors="ignore", inplace=True)
    #
    # if "availability condition" in result:
    #     result["availability condition"] = result["availability condition"].fillna("")
    #
    # for i in (
    #     "chosen",
    #     "chosen but not available",
    #     "chosen thus available",
    #     "not available so not chosen",
    # ):
    #     if i in result.columns and all(result[i] == result[i].astype(int)):
    #         result[i] = result[i].astype(int)
    #
    # for i in ("available",):
    #     if i in result.columns:
    #         j = result.columns.get_loc(i)
    #         if all(result.iloc[:-1, j] == result.iloc[:-1, j].astype(int)):
    #             result.iloc[:-1, j] = result.iloc[:-1, j].astype(int)

    return result



import secrets
import sys

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from numpy.random import SeedSequence, default_rng
from pytest import approx, fixture, mark, raises

import sharrow
from sharrow import Dataset, DataTree, example_data
from sharrow.dataset import from_named_objects


@fixture
def households():
    households = example_data.get_households()
    prng = default_rng(SeedSequence(42))
    households["otaz_idx"] = households["TAZ"] - 1
    households["dtaz_idx"] = prng.choice(np.arange(25), 5000)
    households["timeperiod5"] = prng.choice(np.arange(5), 5000)
    households["timeperiod3"] = np.clip(households["timeperiod5"], 1, 3) - 1
    households["rownum"] = np.arange(len(households))
    return households


@fixture
def skims():
    return example_data.get_skims()


@pytest.mark.parametrize("use_array_maker", [True, False])
def test_shared_data(dataframe_regression, households, skims, use_array_maker: bool):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )

    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "skims.SOV_TIME/base.income",
            "sov_cost_by_income": "skims.HOV3_TIME",
        }
    )
    result = ss.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result, basename="test_shared_data")

    ss_undot = tree.setup_flow(
        {
            "income": "income",
            "sov_time_by_income": "SOV_TIME/income",
            "sov_cost_by_income": "HOV3_TIME",
        }
    )
    result = ss_undot.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result, basename="test_shared_data")

    # names that are not valid Python identifiers
    s2 = tree.setup_flow(
        {
            "income > 10k": "base.income > 10_000",
            "income [up to 10k]": "base.income <= 10_000",
            "sov_time / income": "skims.SOV_TIME/base.income",
            "log1p(sov_cost_by_income)": "log1p(skims.HOV3_TIME)",
        }
    )
    result2 = s2.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result2, basename="test_shared_data_2")


@pytest.mark.parametrize("use_array_maker", [True, False])
def test_subspace_fallbacks(
    dataframe_regression, households, skims, use_array_maker: bool
):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )
    tree.subspace_fallbacks["df"] = ["base", "skims"]

    flow1 = tree.setup_flow(
        {
            "income": "df['income']",
            "sov_time_by_income": "df['SOV_TIME']/df['income']",
            "sov_cost_by_income": "df['HOV3_TIME']",
        }
    )
    result1 = flow1.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result1, basename="test_shared_data")

    flow2 = tree.setup_flow(
        {
            "income": "income",
            "sov_time_by_income": "SOV_TIME/income",
            "sov_cost_by_income": "HOV3_TIME",
        }
    )
    result2 = flow2.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result2, basename="test_shared_data")

    # names that are not valid Python identifiers
    flow3 = tree.setup_flow(
        {
            "income > 10k": "df.income > 10_000",
            "income [up to 10k]": "df.income <= 10_000",
            "sov_time / income": "df.SOV_TIME/df.income",
            "log1p(sov_cost_by_income)": "log1p(df.HOV3_TIME)",
        }
    )
    result3 = flow3.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result3, basename="test_shared_data_2")


@pytest.mark.parametrize("use_array_maker", [True, False])
def test_shared_data_reversible(
    dataframe_regression, households, skims, use_array_maker: bool
):
    tree = DataTree(
        base=households,
        odt_skims=skims,
        dot_skims=skims,
        relationships=(
            "base.otaz_idx->odt_skims.otaz",
            "base.dtaz_idx->odt_skims.dtaz",
            "base.timeperiod5->odt_skims.time_period",
            "base.otaz_idx->dot_skims.dtaz",
            "base.dtaz_idx->dot_skims.otaz",
            "base.timeperiod5->dot_skims.time_period",
        ),
    )

    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "odt_skims.SOV_TIME/base.income",
            "round_trip_hov3_time": "dot_skims.HOV3_TIME + odt_skims.HOV3_TIME",
            "double_hov3_time": "odt_skims.HOV3_TIME * 2",
        }
    )
    result = ss.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result, basename="test_shared_data_reversible")
    with raises(AssertionError):
        pd.testing.assert_series_equal(
            result["round_trip_hov3_time"],
            result["double_hov3_time"],
        )


@pytest.mark.parametrize("use_array_maker", [True, False])
def test_shared_data_reversible_by_label(dataframe_regression, use_array_maker: bool):
    data = example_data.get_data()
    skims = data["skims"]
    households = data["hhs"]

    # households = pd.read_csv("data/households.csv.gz")
    prng = default_rng(SeedSequence(42))
    households["otaz"] = households["TAZ"]
    households["dtaz"] = prng.choice(np.arange(1, 26), 5000)
    households["time5"] = prng.choice(["EA", "AM", "MD", "PM", "EV"], 5000)
    households["time3"] = prng.choice(["AM", "MD", "PM"], 5000)

    tree = DataTree(
        base=households,
        odt_skims=skims.rename({"otaz": "ptaz", "dtaz": "ataz"}),
        dot_skims=skims.rename({"otaz": "ataz", "dtaz": "ptaz"}),
        relationships=(
            "base.otaz @ odt_skims.ptaz",
            "base.dtaz @ odt_skims.ataz",
            "base.time5 @ odt_skims.time_period",
            "base.otaz @ dot_skims.ptaz",
            "base.dtaz @ dot_skims.ataz",
            "base.time5 @ dot_skims.time_period",
        ),
    )

    tree.digitize_relationships(inplace=True)
    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "odt_skims.SOV_TIME/base.income",
            "round_trip_hov3_time": "dot_skims.HOV3_TIME + odt_skims.HOV3_TIME",
            "double_hov3_time": "odt_skims.HOV3_TIME * 2",
        },
        extra_hash_data=(
            1,
            2,
        ),
    )
    result = ss.load_dataframe(tree, use_array_maker=use_array_maker)
    dataframe_regression.check(result, basename="test_shared_data_reversible")
    with raises(AssertionError):
        pd.testing.assert_series_equal(
            result["round_trip_hov3_time"],
            result["double_hov3_time"],
        )

    dtree = tree.digitize_relationships()
    dss = dtree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "odt_skims.SOV_TIME/base.income",
            "round_trip_hov3_time": "dot_skims.HOV3_TIME + odt_skims.HOV3_TIME",
            "double_hov3_time": "odt_skims.HOV3_TIME * 2",
        },
        extra_hash_data=(
            1,
            2,
            3,
        ),
    )
    dresult = dss.load_dataframe(dtree, use_array_maker=use_array_maker)
    dataframe_regression.check(dresult, basename="test_shared_data_reversible")


@pytest.mark.parametrize("use_array_maker", [False, True])
def test_with_2d_base(dataframe_regression, use_array_maker: bool):
    pytest.importorskip("scipy", minversion="0.16")

    data = example_data.get_data()
    skims = data["skims"]
    households = data["hhs"]

    prng = default_rng(SeedSequence(42))
    households["otaz"] = households["TAZ"]
    households["otaz_idx"] = households["TAZ"] - 1
    households["dtaz"] = prng.choice(np.arange(1, 26), 5000)
    households["timeperiod5"] = prng.choice(np.arange(5), 5000)
    households["timeperiod3"] = np.clip(households["timeperiod5"], 1, 3) - 1
    households["rownum"] = np.arange(len(households))
    households["time5"] = prng.choice(["EA", "AM", "MD", "PM", "EV"], 5000)
    households["time3"] = prng.choice(["AM", "MD", "PM"], 5000)

    blank = from_named_objects(households.index, skims["dtaz"])
    assert sorted(blank.coords) == ["HHID", "dtaz"]
    assert blank.coords["HHID"].dims == ("HHID",)
    assert blank.coords["dtaz"].dims == ("dtaz",)

    tree = DataTree(
        root_node_name="base",
        base=blank,
        hh=households,
        odt_skims=skims.rename({"otaz": "ptaz", "dtaz": "ataz"}),
        dot_skims=skims.rename({"otaz": "ataz", "dtaz": "ptaz"}),
        relationships=(
            "base.HHID @ hh.HHID",
            "base.dtaz @ odt_skims.ataz",
            "base.dtaz @ dot_skims.ataz",
            "hh.otaz @ odt_skims.ptaz",
            "hh.time5 @ odt_skims.time_period",
            "hh.otaz @ dot_skims.ptaz",
            "hh.time5 @ dot_skims.time_period",
        ),
        force_digitization=True,
    )

    ss = tree.setup_flow(
        {
            "income": "hh.income",
            "sov_time_by_income": "odt_skims.SOV_TIME/hh.income",
            "round_trip_hov3_time": "dot_skims.HOV3_TIME + odt_skims.HOV3_TIME",
            "double_hov3_time": "odt_skims.HOV3_TIME * 2",
            "a_trip_hov3_time": "dot_skims.HOV3_TIME",
            "b_trip_hov3_time": "odt_skims.HOV3_TIME",
        }
    )
    result = ss.load_dataarray(tree, use_array_maker=use_array_maker)
    assert result.dims == ("HHID", "dtaz", "expressions")
    assert result.shape == (5000, 25, 6)
    result = result.to_dataset("expressions").to_dataframe()
    dataframe_regression.check(result.iloc[::83], basename="test_with_2d_base")

    dot_result = ss._load(tree, as_dataarray=True, dot=np.ones(6))
    assert dot_result.dims == (
        "HHID",
        "dtaz",
    )
    assert dot_result.shape == (
        5000,
        25,
    )

    check_vs = np.dot(result, np.ones([6])).reshape(5000, 25)
    np.testing.assert_array_almost_equal(check_vs, dot_result.to_numpy())


@pytest.mark.parametrize("use_array_maker", [True, False])
def test_mixed_dtypes(dataframe_regression, households, skims, use_array_maker: bool):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )

    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "skims.SOV_TIME/base.income",
            "sov_time_by_workers": "np.where(base.workers > 0, skims.SOV_TIME / base.workers, 0)",
        }
    )
    result = ss.load_dataframe(tree, dtype=np.float32, use_array_maker=use_array_maker)
    dataframe_regression.check(result, basename="test_mixed_dtypes")

    ss_undot = tree.setup_flow(
        {
            "income": "income",
            "sov_time_by_income": "SOV_TIME/income",
            "sov_time_by_workers": "np.where(workers > 0, SOV_TIME / workers, 0)",
        }
    )
    result = ss_undot.load_dataframe(
        tree, dtype=np.float32, use_array_maker=use_array_maker
    )
    dataframe_regression.check(result, basename="test_mixed_dtypes")


@pytest.mark.parametrize("use_array_maker", [True, False])
def test_tuple_slice(dataframe_regression, households, skims, use_array_maker):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
        ),
    )
    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_md": "skims[('SOV_TIME', 'MD')]",
        }
    )
    result = ss.load_dataframe(tree, dtype=np.float32, use_array_maker=use_array_maker)
    dataframe_regression.check(result, basename="test_tuple_slice")


@pytest.mark.parametrize("use_array_maker", [True, False])
def test_isin(dataframe_regression, households, skims, use_array_maker: bool):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
        ),
        extra_vars={
            "twenty_two_hundred": 2200,
            "sixteen_five": 16500,
        },
    )
    ss = tree.setup_flow(
        {
            "income": "base.income",
            "income2": "base.income.isin([361000, 197000])",
            "income3": "base.income.isin([twenty_two_hundred, sixteen_five])",
            "income4": "base.income.isin((twenty_two_hundred, 197000))",
        }
    )
    result = ss.load_dataframe(tree, dtype=np.float32, use_array_maker=use_array_maker)
    dataframe_regression.check(result, basename="test_isin")


def _get_target(q, token):
    skims_ = Dataset.shm.from_shared_memory(token)
    q.put(skims_.SOV_TIME.sum())


@mark.skipif(
    sys.version_info < (3, 8), reason="shared memory requires python3.8 or higher"
)
def test_shared_memory(skims):
    token = "skims" + secrets.token_hex(5)

    skims_2 = skims.shm.to_shared_memory(token)
    target = skims.SOV_TIME.sum()
    assert skims_2.SOV_TIME.sum() == target

    # reconstruct in same process
    skims_3 = Dataset.shm.from_shared_memory(token)
    assert skims_3.SOV_TIME.sum() == target

    # reconstruct in different process
    from multiprocessing import Process, Queue

    q = Queue()
    p = Process(target=_get_target, args=(q, token))
    p.start()
    p.join()
    assert q.get() == target


def test_relationship_init():
    r = sharrow.Relationship.from_string("Aa.bb -> Cc.dd")
    assert r.parent_data == "Aa"
    assert r.parent_name == "bb"
    assert r.child_data == "Cc"
    assert r.child_name == "dd"
    assert r.indexing == "position"

    r = sharrow.Relationship.from_string("Ee.ff @ Gg.hh")
    assert r.parent_data == "Ee"
    assert r.parent_name == "ff"
    assert r.child_data == "Gg"
    assert r.child_name == "hh"
    assert r.indexing == "label"


def test_replacement_filters(dataframe_regression, households, skims):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )

    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "skims.SOV_TIME/base.income",
            "sov_cost_by_income": "skims.HOV3_TIME",
        }
    )
    result = ss._load(tree, as_dataframe=True)
    dataframe_regression.check(result, basename="test_shared_data")

    households_malformed = households.rename(columns={"income": "jncome"})
    tree_malformed = ss.tree.replace_datasets(base=households_malformed)
    with raises(KeyError, match=".*income.*"):
        ss._load(tree_malformed, as_dataframe=True)

    def rename_jncome(x):
        return x.rename({"jncome": "income"})

    ss.tree.replacement_filters["base"] = rename_jncome

    result = ss._load(
        ss.tree.replace_datasets(base=households_malformed), as_dataframe=True
    )
    dataframe_regression.check(result, basename="test_shared_data")


def test_name_in_wrong_subspace(dataframe_regression, households, skims):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )

    with raises(KeyError):
        tree.setup_flow(
            {
                "income": "base.income",
                "sov_time_by_income": "base.SOV_TIME/base.income",
                "sov_cost_by_income": "base.HOV3_TIME",
            }
        )

    tree = DataTree(
        base=households,
        od_skims=skims.drop_dims("time_period"),
        odt_skims=skims,
        relationships=(
            "base.otaz_idx->od_skims.otaz",
            "base.dtaz_idx->od_skims.dtaz",
            "base.otaz_idx->odt_skims.otaz",
            "base.dtaz_idx->odt_skims.dtaz",
            "base.timeperiod5->odt_skims.time_period",
        ),
    )
    with raises(KeyError):
        tree.setup_flow(
            {
                "income": "base.income",
                "SOV_TIMEbyINC": "SOV_TIME/base.income",
                "SOV_TIMEbyINC1": "od_skims.SOV_TIME/base.income",
                "SOV_TIME": "od_skims.SOV_TIME",
                "HOV3_TIME": "od_skims.HOV3_TIME",
                "SOV_TIME_t": "odt_skims.SOV_TIME",
                "HOV3_TIME_t": "odt_skims.HOV3_TIME",
            }
        )
    with raises(KeyError):
        tree.setup_flow(
            {
                "income": "base.income",
                "SOV_TIME": "od_skims.SOV_TIME",
                "HOV3_TIME": "od_skims.HOV3_TIME",
                "SOV_TIME_t": "odt_skims.SOV_TIME",
                "HOV3_TIME_t": "odt_skims.HOV3_TIME",
            }
        )

    tree.subspace_fallbacks["od_skims"] = ["odt_skims"]

    ss_undot = tree.setup_flow(
        {
            "income": "income",
            "sov_time_by_income": "od_skims.SOV_TIME/income",
            "sov_cost_by_income": "od_skims.HOV3_TIME",
        }
    )
    result = ss_undot._load(tree, as_dataframe=True)
    dataframe_regression.check(result, basename="test_shared_data")


def test_shared_data_encoded(dataframe_regression, households, skims):
    households = sharrow.dataset.construct(households).digital_encoding.set(
        "income",
        bitwidth=32,
        scale=0.001,
    )

    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )

    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "skims.SOV_TIME/base.income",
            "sov_cost_by_income": "skims.HOV3_TIME",
        },
        extra_hash_data=("income_encoded",),
    )
    result = ss._load(tree, as_dataframe=True)
    dataframe_regression.check(result, basename="test_shared_data")


def test_dict_encoded(dataframe_regression, skims):
    pairs = pd.DataFrame({"orig": [0, 0, 0, 1, 1, 1], "dest": [0, 1, 2, 0, 1, 2]})
    skims1 = skims.digital_encoding.set("WLK_LOC_WLK_FAR", bitwidth=8, by_dict=True)
    tree1 = DataTree(
        base=pairs,
        skims=skims1,
        relationships=(
            "base.orig -> skims.otaz",
            "base.dest -> skims.dtaz",
        ),
    )
    flow1 = tree1.setup_flow(
        {
            "d1": 'skims["WLK_LOC_WLK_FAR", "AM"]',
            "d2": 'skims["WLK_LOC_WLK_FAR", "AM"]**2',
        }
    )
    arr1 = flow1.load_dataframe()
    dataframe_regression.check(arr1)


def test_joint_dict_encoded(dataframe_regression, skims):
    pairs = pd.DataFrame({"orig": [0, 0, 0, 1, 1, 1], "dest": [0, 1, 2, 0, 1, 2]})
    skims1 = skims.digital_encoding.set(
        "WLK_LOC_WLK_FAR",
        "WLK_LOC_WLK_BOARDS",
        "WLK_LOC_WLK_IWAIT",
        "WLK_LOC_WLK_WAIT",
        joint_dict=True,
    )
    skims1 = skims1.digital_encoding.set(
        ["DISTBIKE", "DISTWALK"],
        joint_dict="jointWB",
    )
    tree1 = DataTree(
        base=pairs,
        skims=skims1,
        rskims=skims1,
        relationships=(
            "base.orig -> skims.otaz",
            "base.dest -> skims.dtaz",
            "base.orig -> rskims.dtaz",
            "base.dest -> rskims.otaz",
        ),
    )
    flow1 = tree1.setup_flow(
        {
            "f1": 'skims["WLK_LOC_WLK_FAR", "AM"]',
            "f2": 'skims["WLK_LOC_WLK_FAR", "AM"]**2',
            "w1": "skims.DISTWALK",
            "w2": 'skims.reverse("DISTWALK")',
            "w3": "rskims.DISTWALK",
            "x1": "skims.DIST",
            "x2": 'skims.reverse("DIST")',
        },
        extra_hash_data=("joint",),
    )
    arr1 = flow1.load_dataframe()
    dataframe_regression.check(arr1)


def test_isin_and_between(dataframe_regression):
    data = example_data.get_data()
    persons = data["persons"]

    tree = DataTree(
        base=persons,
        extra_vars={
            "pt1": 1,
            "pt5": 5,
            "pt34": [3, 4],
        },
    )

    ss = tree.setup_flow(
        {
            "pt": "base.ptype",
            "pt_in_15": "base.ptype.isin([pt1,pt5])",
            "pt_in_34": "base.ptype.isin(pt34)",
            "pt_tween_15": "base.ptype.between(pt1,pt5)",
            "pt_tween_25": "base.ptype.between(2,pt5)",
            "pt_tween_35": "base.ptype.between(3,5)",
        }
    )
    result = ss.load_dataframe(tree)
    pd.testing.assert_series_equal(
        result["pt"].isin([1, 5]).astype(np.float32),
        result["pt_in_15"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].isin([3, 4]).astype(np.float32),
        result["pt_in_34"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].between(1, 5).astype(np.float32),
        result["pt_tween_15"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].between(2, 5).astype(np.float32),
        result["pt_tween_25"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].between(3, 5).astype(np.float32),
        result["pt_tween_35"],
        check_names=False,
    )
    dataframe_regression.check(result)

    # test masking
    mask = (persons.index % 2) == 0

    result = ss.load_dataframe(tree, mask=mask)
    pd.testing.assert_series_equal(
        result["pt"].isin([1, 5]).astype(np.float32).where(mask, np.nan),
        result["pt_in_15"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].isin([3, 4]).astype(np.float32).where(mask, np.nan),
        result["pt_in_34"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].between(1, 5).astype(np.float32).where(mask, np.nan),
        result["pt_tween_15"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].between(2, 5).astype(np.float32).where(mask, np.nan),
        result["pt_tween_25"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        result["pt"].between(3, 5).astype(np.float32).where(mask, np.nan),
        result["pt_tween_35"],
        check_names=False,
    )


def test_nested_where(dataframe_regression):
    data = example_data.get_data()
    base = persons = data["persons"]

    tree = DataTree(
        base=persons,
        extra_vars={
            "pt1": 1,
            "pt5": 5,
            "pt34": [3, 4],
        },
    )

    ss = tree.setup_flow(
        {
            "pt": "base.ptype",
            "pt_shifted_1": "np.where(base.ptype<3, np.where(base.ptype<2, base.ptype*100, 0), base.ptype)",
            "pt_shifted_2": "np.where(base.ptype<3, np.where(base.ptype<2, base.ptype*100, 0), 0)",
            "pt_shifted_3": "np.where(base.ptype<3, 0, np.where(base.ptype>4, base.ptype*100, 0))",
            "pt_shifted_4": "np.where(base.ptype<3, base.ptype, np.where(base.ptype>4, base.ptype*100, 0))",
            "pt_shifted_5": "np.where(base.ptype<3, base.ptype, np.where(base.ptype>4, base.ptype*100, base.ptype))",
            "pt_shifted_6": "np.where(base.ptype<3, 0, np.where(base.ptype>4, base.ptype*100, base.ptype))",
        }
    )
    result = ss.load_dataframe(tree)
    pd.testing.assert_series_equal(
        pd.Series(
            np.where(
                base.ptype < 3,
                np.where(base.ptype < 2, base.ptype * 100, 0),
                base.ptype,
            ).astype(np.float32),
        ),
        result["pt_shifted_1"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        pd.Series(
            np.where(
                base.ptype < 3, np.where(base.ptype < 2, base.ptype * 100, 0), 0
            ).astype(np.float32),
        ),
        result["pt_shifted_2"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        pd.Series(
            np.where(
                base.ptype < 3, 0, np.where(base.ptype > 4, base.ptype * 100, 0)
            ).astype(np.float32),
        ),
        result["pt_shifted_3"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        pd.Series(
            np.where(
                base.ptype < 3,
                base.ptype,
                np.where(base.ptype > 4, base.ptype * 100, 0),
            ).astype(np.float32),
        ),
        result["pt_shifted_4"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        pd.Series(
            np.where(
                base.ptype < 3,
                base.ptype,
                np.where(base.ptype > 4, base.ptype * 100, base.ptype),
            ).astype(np.float32),
        ),
        result["pt_shifted_5"],
        check_names=False,
    )
    pd.testing.assert_series_equal(
        pd.Series(
            np.where(
                base.ptype < 3,
                0,
                np.where(base.ptype > 4, base.ptype * 100, base.ptype),
            ).astype(np.float32),
        ),
        result["pt_shifted_6"],
        check_names=False,
    )
    dataframe_regression.check(result)


def test_isna():
    data = example_data.get_data()
    data["hhs"].loc[data["hhs"].income > 200000, "income"] = np.nan
    tree = DataTree(
        base=data["hhs"],
    )
    ss = tree.setup_flow(
        {
            "missing_income": "((income < 0) | income.isna())",
            "income_is_na": "income.isna()",
        }
    )
    result = ss.load()
    assert result[0, 0] == 1
    assert result[0, 1] == 1
    assert result[:, 0].sum() == 188
    assert result[:, 1].sum() == 188

    qf = pd.DataFrame({"MixedVals": ["a", "", None, np.nan]})
    tree2 = DataTree(
        base=qf,
    )
    qf = tree2.setup_flow(
        {
            "MixedVals_is_na": "MixedVals.isna()",
        }
    )
    result = qf.load()
    assert result == approx(np.asarray([[0, 0, 1, 1]]).T)


def test_get(dataframe_regression, households, skims):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )

    flow1 = tree.setup_flow(
        {
            "income": "base.get('income', 0) + base.get('missing_one', 0)",
            "sov_time_by_income": "skims.SOV_TIME/base.get('income', 0)",
            "missing_data": "base.get('missing_data', -1)",
            "missing_skim": "skims.get('missing_core', -2)",
            "sov_time_by_income_2": "skims.get('SOV_TIME')/base.income",
            "sov_cost_by_income_2": "skims.get('HOV3_TIME', 999)",
        },
    )
    result = flow1._load(tree, as_dataframe=True)
    dataframe_regression.check(result)

    tree_plus = DataTree(
        base=households.assign(missing_one=1.0),
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )
    flow2 = tree_plus.setup_flow(flow1.defs)
    result = flow2._load(tree_plus, as_dataframe=True)
    dataframe_regression.check(result.eval("income = income-1"))
    assert flow2.flow_hash != flow1.flow_hash

    tree.subspace_fallbacks["df"] = ["base"]
    flow3 = tree.setup_flow(
        {
            "income": "base.get('income', 0)",
            "sov_time_by_income": "skims.SOV_TIME/df.get('income', 0)",
            "missing_data": "df.get('missing_data', -1)",
            "missing_skim": "skims.get('missing_core', -2)",
            "sov_time_by_income_2": "skims.get('SOV_TIME')/df.income",
            "sov_cost_by_income_2": "skims.get('HOV3_TIME', 999)",
        },
    )
    result = flow3._load(tree, as_dataframe=True)
    dataframe_regression.check(result)

    flow4 = tree.setup_flow(
        {
            "income": "base.get('income', default=0) + df.get('missing_one', 0)",
            "sov_time_by_income": "skims.SOV_TIME/base.get('income', default=0)",
            "missing_data": "base.get('missing_data', default=-1)",
            "missing_skim": "skims.get('missing_core', default=-2)",
            "sov_time_by_income_2": "skims.get('SOV_TIME', default=0)/base.income",
            "sov_cost_by_income_2": "skims.get('HOV3_TIME', default=999)",
        },
    )
    result = flow4._load(tree, as_dataframe=True)
    assert flow4.flow_hash != flow1.flow_hash
    dataframe_regression.check(result)

    # test get when inside another function
    flow5 = tree.setup_flow(
        {
            "income": "np.power(base.get('income', default=0) + df.get('missing_one', 0), 1)",
            "sov_time_by_income": "skims.SOV_TIME/np.power(base.get('income', default=0), 1)",
            "missing_data": "np.where(np.isnan(df.get('missing_data', default=1)), 0, df.get('missing_data', default=-1))",  # noqa: E501
            "missing_skim": "(np.where(np.isnan(df.get('num_escortees', np.nan)), -2 , df.get('num_escortees', np.nan)))",  # noqa: E501
            "sov_time_by_income_2": "skims.get('SOV_TIME', default=0)/base.income",
            "sov_cost_by_income_2": "skims.get('HOV3_TIME', default=999)",
        },
    )
    result = flow5._load(tree, as_dataframe=True)
    assert "__skims__HOV3_TIME:True" in flow5._optional_get_tokens
    assert "__df__missing_data:False" in flow5._optional_get_tokens
    assert "__df__num_escortees:False" in flow5._optional_get_tokens
    dataframe_regression.check(result)


def test_get_native():
    data = example_data.get_data()
    skims = data["skims"]
    households = data["hhs"]

    prng = default_rng(SeedSequence(42))
    households["otaz"] = households["TAZ"]
    households["otaz_idx"] = households["TAZ"] - 1
    households["dtaz"] = prng.choice(np.arange(1, 26), 5000)
    households["timeperiod5"] = prng.choice(np.arange(5), 5000)
    households["timeperiod3"] = np.clip(households["timeperiod5"], 1, 3) - 1
    households["rownum"] = np.arange(len(households))
    households["time5"] = prng.choice(["EA", "AM", "MD", "PM", "EV"], 5000)
    households["time3"] = prng.choice(["AM", "MD", "PM"], 5000)

    blank = from_named_objects(households.index, skims["dtaz"])
    assert sorted(blank.coords) == ["HHID", "dtaz"]
    assert blank.coords["HHID"].dims == ("HHID",)
    assert blank.coords["dtaz"].dims == ("dtaz",)

    tree = DataTree(
        root_node_name="base",
        base=blank,
        hh=households,
        odt_skims=skims.rename({"otaz": "ptaz", "dtaz": "ataz"}),
        dot_skims=skims.rename({"otaz": "ataz", "dtaz": "ptaz"}),
        relationships=(
            "base.HHID @ hh.HHID",
            "base.dtaz @ odt_skims.ataz",
            "base.dtaz @ dot_skims.ataz",
            "hh.otaz @ odt_skims.ptaz",
            "hh.time5 @ odt_skims.time_period",
            "hh.otaz @ dot_skims.ptaz",
            "hh.time5 @ dot_skims.time_period",
        ),
        force_digitization=True,
    )

    i = tree.get("missingname", 123)
    assert i.dims == ("HHID", "dtaz")
    assert i.shape == (5000, 25)
    assert (i == 123).all()

    i = tree["income"]
    assert isinstance(i, xr.DataArray)
    assert i.dims == ("HHID", "dtaz")
    assert i.shape == (5000, 25)
    assert i.std("dtaz").max() == 0

    i = tree.get(["income", "PERSONS"], broadcast=False)
    assert isinstance(i, xr.Dataset)
    assert "HHID" in i.coords
    assert i.dims == {"HHID": 5000}

    i = tree.get("income", broadcast=False)
    assert "HHID" in i.coords
    assert i.dims == ("HHID",)
    assert i.shape == (5000,)

    i = tree.get("income", broadcast=False, coords=False)
    assert not i.coords
    assert i.dims == ("HHID",)
    assert i.shape == (5000,)

    i = tree.get("odt_skims.DIST")
    assert i.dims == ("HHID", "dtaz")
    assert i.shape == (5000, 25)
    assert (tree.get_expr("odt_skims.DIST", allow_native=False) == i.load()).all()

    with raises(KeyError):
        tree.get("xxxx.income")
    with raises(KeyError):
        tree.get("base.xxxx")
    with raises(KeyError):
        tree.get("xxxx")
    with raises(KeyError):
        tree.get("base.DIST")


def test_streaming(households, skims):
    tree = DataTree(
        base=households,
        skims=skims,
        relationships=(
            "base.otaz_idx->skims.otaz",
            "base.dtaz_idx->skims.dtaz",
            "base.timeperiod5->skims.time_period",
        ),
    )
    ss = tree.setup_flow(
        {
            "income": "base.income",
            "sov_time_by_income": "skims.SOV_TIME/base.income",
            "sov_cost_by_income": "skims.HOV3_TIME",
        }
    )
    result = ss.load(tree)
    streamer = ss.init_streamer(tree)
    for i in range(len(households)):
        assert result[i] == approx(streamer(i))


def test_streaming_2d(households, skims):
    blank = from_named_objects(households.index, skims["dtaz"])
    assert sorted(blank.coords) == ["HHID", "dtaz"]
    assert blank.coords["HHID"].dims == ("HHID",)
    assert blank.coords["dtaz"].dims == ("dtaz",)

    tree = DataTree(
        root_node_name="base",
        base=blank,
        hh=households,
        odt_skims=skims.rename({"otaz": "ptaz", "dtaz": "ataz"}),
        dot_skims=skims.rename({"otaz": "ataz", "dtaz": "ptaz"}),
        relationships=(
            "base.HHID @ hh.HHID",
            "base.dtaz @ odt_skims.ataz",
            "base.dtaz @ dot_skims.ataz",
            "hh.otaz_idx -> odt_skims.ptaz",
            "hh.timeperiod5 -> odt_skims.time_period",
            "hh.otaz_idx -> dot_skims.ptaz",
            "hh.timeperiod5 -> dot_skims.time_period",
        ),
        force_digitization=True,
    )

    ss = tree.setup_flow(
        {
            "income": "hh.income",
            "sov_time_by_income": "odt_skims.SOV_TIME/hh.income",
            "round_trip_hov3_time": "dot_skims.HOV3_TIME + odt_skims.HOV3_TIME",
            "double_hov3_time": "odt_skims.HOV3_TIME * 2",
            "a_trip_hov3_time": "dot_skims.HOV3_TIME",
            "b_trip_hov3_time": "odt_skims.HOV3_TIME",
        }
    )
    result = ss.load_dataarray(tree)
    assert result.dims == ("HHID", "dtaz", "expressions")
    assert result.shape == (5000, 25, 6)
    result = result.to_numpy()
    streamer = ss.init_streamer(tree)
    assert streamer(0).shape == (25, 6)
    for i in range(len(households)):
        assert (result[i] == streamer(i)).all()


@pytest.mark.parametrize("force_digitization", [True, False])
@pytest.mark.parametrize("engine", ["sharrow", "numexpr", "python", None])
def test_eval_1d_root(force_digitization: bool, engine: str | None):
    data = example_data.get_data()
    skim = data["skims"]
    hh = data["hhs"]

    prng = default_rng(SeedSequence(42))
    hh["otaz"] = hh["TAZ"]
    hh["otaz_idx"] = hh["TAZ"] - 1
    hh["dtaz"] = prng.choice(np.arange(1, 26), 5000)
    hh["timeperiod5"] = prng.choice(np.arange(5), 5000)
    hh["timeperiod3"] = np.clip(hh["timeperiod5"], 1, 3) - 1
    hh["rownum"] = np.arange(len(hh))
    hh["time5"] = prng.choice(["EA", "AM", "MD", "PM", "EV"], 5000)
    hh["time3"] = prng.choice(["AM", "MD", "PM"], 5000)
    hh["n_tours"] = prng.choice([0, 1, 2, 3], 5000)

    n_tours = hh["n_tours"].sum()
    tours = pd.DataFrame(
        {
            "HHID": np.repeat(hh.index, hh["n_tours"]),
            "tour_id": np.arange(1, n_tours + 1),
            "dtaz": prng.choice(np.arange(1, 26), n_tours),
            "tourtime": prng.choice(["EA", "AM", "MD", "PM", "EV"], n_tours),
        }
    ).set_index("tour_id")
    skim.load()

    tree = DataTree(
        tour=tours,
        hh=hh,
        odt_skims=skim,
        dot_skims=skim,
        relationships=(
            "tour.HHID @ hh.HHID",
            "tour.dtaz @ odt_skims.dtaz",
            "tour.dtaz @ dot_skims.otaz",
            "hh.otaz_idx -> odt_skims.otaz",
            "tour.tourtime @ odt_skims.time_period",
            "hh.otaz_idx -> dot_skims.dtaz",
            "tour.tourtime @ dot_skims.time_period",
        ),
        force_digitization=force_digitization,
    )

    sov_time = tree["SOV_TIME"]
    assert sov_time.shape == (7563,)
    assert sov_time.dims == ("tour_id",)
    assert sov_time[:3].values == approx([4.94, 5.02, 3.61])
    assert sov_time[-3:].values == approx([7.17, 5.53, 2.65])

    sov_time_by_income = tree["SOV_TIME"] / tree["income"]
    assert sov_time_by_income.shape == (7563,)
    assert sov_time_by_income.dims == ("tour_id",)
    assert sov_time_by_income[:3].values == approx(
        [8.34177652e-05, 8.47686589e-05, 1.64090904e-03]
    )
    assert sov_time_by_income[-3:].values == approx(
        [5.47328250e-04, 5.36893224e-05, 2.57281563e-05]
    )

    t = tree.eval("SOV_TIME", engine=engine)
    assert "expressions" in t.coords, f"missing expression coords with engine={engine}"
    assert t.coords["expressions"] == "SOV_TIME"
    xr.testing.assert_allclose(t.drop_vars("expressions"), sov_time)

    t = tree.eval("SOV_TIME / income", engine=engine)
    assert "expressions" in t.coords, f"missing expression coords with engine={engine}"
    assert t.coords["expressions"] == "SOV_TIME / income"
    xr.testing.assert_allclose(t.drop_vars("expressions"), sov_time_by_income)

    dot_time = tree["dot_skims.SOV_TIME"]
    assert dot_time.shape == (7563,)
    assert dot_time.dims == ("tour_id",)
    assert dot_time[:3].values == approx([5.3, 4.69, 3.58])
    assert dot_time[-3:].values == approx([7.36, 5.67, 3.08])

    dot_time_by_income = tree["dot_skims.SOV_TIME"] / tree["hh.income"]
    assert dot_time_by_income.shape == (7563,)
    assert dot_time_by_income.dims == ("tour_id",)
    assert dot_time_by_income[:3].values == approx(
        [8.94967948e-05, 7.91962185e-05, 1.62727269e-03]
    )
    assert dot_time_by_income[-3:].values == approx(
        [5.61832071e-04, 5.50485444e-05, 2.99029119e-05]
    )

    t = tree.eval("dot_skims.SOV_TIME", engine=engine)
    assert "expressions" in t.coords, f"missing expression coords with engine={engine}"
    assert t.coords["expressions"] == "dot_skims.SOV_TIME"
    xr.testing.assert_allclose(t.drop_vars("expressions"), dot_time)

    t = tree.eval("dot_skims.SOV_TIME / income", engine=engine)
    assert "expressions" in t.coords, f"missing expression coords with engine={engine}"
    assert t.coords["expressions"] == "dot_skims.SOV_TIME / income"
    xr.testing.assert_allclose(t.drop_vars("expressions"), dot_time_by_income)

    tm = tree.eval_many(
        {"SOV_TIME": "SOV_TIME", "SOV_TIME_by_income": "SOV_TIME/income"},
        engine=engine,
        result_type="dataarray",
    )
    assert tm.dims == ("tour_id", "expressions")
    assert tm.coords["expressions"].values.tolist() == [
        "SOV_TIME",
        "SOV_TIME_by_income",
    ]
    xr.testing.assert_allclose(
        tm.drop_vars(["expressions", "source"]),
        xr.concat(
            [
                sov_time.expand_dims("expressions", -1),
                sov_time_by_income.expand_dims("expressions", -1),
            ],
            "expressions",
        ),
    )

    tm = tree.eval_many(
        {"SOV_TIME": "SOV_TIME", "SOV_TIME_by_income": "SOV_TIME/income"},
        engine=engine,
        result_type="dataset",
    )
    assert tm.sizes == {"tour_id": 7563}
    xr.testing.assert_allclose(
        tm,
        xr.merge(
            [
                sov_time.rename("SOV_TIME"),
                sov_time_by_income.rename("SOV_TIME_by_income"),
            ]
        ),
    )


@pytest.mark.parametrize("force_digitization", [True, False])
@pytest.mark.parametrize("engine", ["sharrow", "numexpr", "python", None])
def test_eval_2d_root(
    force_digitization: bool, dataframe_regression, engine: str | None
):
    pytest.importorskip("scipy", minversion="0.16")

    data = example_data.get_data()
    skims = data["skims"]
    households = data["hhs"]

    prng = default_rng(SeedSequence(42))
    households["otaz"] = households["TAZ"]
    households["otaz_idx"] = households["TAZ"] - 1
    households["dtaz"] = prng.choice(np.arange(1, 26), 5000)
    households["timeperiod5"] = prng.choice(np.arange(5), 5000)
    households["timeperiod3"] = np.clip(households["timeperiod5"], 1, 3) - 1
    households["rownum"] = np.arange(len(households))
    households["time5"] = prng.choice(["EA", "AM", "MD", "PM", "EV"], 5000)
    households["time3"] = prng.choice(["AM", "MD", "PM"], 5000)

    blank = from_named_objects(households.index, skims["dtaz"])
    assert sorted(blank.coords) == ["HHID", "dtaz"]
    assert blank.coords["HHID"].dims == ("HHID",)
    assert blank.coords["dtaz"].dims == ("dtaz",)

    skims.load()

    tree = DataTree(
        root_node_name="base",
        base=blank,
        hh=households,
        odt_skims=skims.rename({"otaz": "ptaz", "dtaz": "ataz"}),
        dot_skims=skims.rename({"otaz": "ataz", "dtaz": "ptaz"}),
        relationships=(
            "base.HHID @ hh.HHID",
            "base.dtaz @ odt_skims.ataz",
            "base.dtaz @ dot_skims.ataz",
            "hh.otaz @ odt_skims.ptaz",
            "hh.time5 @ odt_skims.time_period",
            "hh.otaz @ dot_skims.ptaz",
            "hh.time5 @ dot_skims.time_period",
        ),
        force_digitization=force_digitization,
    )

    exprs = {
        "income": "income",
        "sov_time_by_income": "SOV_TIME/income",
        "a_trip_hov_time": "log(exp(HOV2_TIME / 0.5) + exp(HOV3_TIME / 0.5)) * 0.5",
        "round_trip_hov3_time": "dot_skims.HOV3_TIME + odt_skims.HOV3_TIME",
    }

    arrays = {
        k: tree.eval(v, engine=engine, name=k)
        .drop_vars("expressions")
        .assign_attrs(expression=v)
        for (k, v) in exprs.items()
    }
    # result = xr.concat(list(arrays.values()), "expressions")
    result = xr.Dataset(arrays).to_dataframe()
    dataframe_regression.check(result.iloc[::83], basename="test_eval_2d_root")

    result2 = tree.eval_many(exprs, engine=engine, result_type="dataset").to_dataframe()
    dataframe_regression.check(result2.iloc[::83], basename="test_eval_2d_root")



import pandas as pd
import pytest
import xarray as xr

from sharrow.dataset import construct


@pytest.fixture
def person_dataset() -> xr.Dataset:
    """Sample persons dataset with dummy data."""
    df = pd.DataFrame(
        {
            "Income": [45, 88, 56, 15, 71],
            "Name": ["Andre", "Bruce", "Carol", "David", "Eugene"],
            "Age": [14, 25, 55, 8, 21],
            "WorkMode": ["Car", "Bus", "Car", "Car", "Walk"],
            "household_id": [11, 11, 22, 22, 33],
        },
        index=pd.Index([441, 445, 552, 556, 934], name="person_id"),
    )
    df["WorkMode"] = df["WorkMode"].astype("category")
    return construct(df)


@pytest.fixture
def household_dataset() -> xr.Dataset:
    """Sample household dataset with dummy data."""
    df = pd.DataFrame(
        {
            "n_cars": [1, 2, 1],
        },
        index=pd.Index([11, 22, 33], name="household_id"),
    )
    return construct(df)


@pytest.fixture
def tours_dataset() -> xr.Dataset:
    """Sample tours dataset with dummy data."""
    df = pd.DataFrame(
        {
            "TourMode": ["Car", "Bus", "Car", "Car", "Walk"],
            "person_id": [441, 445, 552, 556, 934],
            "origin": [101, 102, 103, 101, 103],
            "destination": [102, 103, 103, 101, 102],
            "time_period": ["AM", "MD", "AM", "AM", "MD"],
            "origin_idx": [0, 1, 2, 0, 2],
            "destination_idx": [1, 2, 2, 0, 1],
            "time_period_alt": ["AM", "MD", "AM", "AM", "MD"],
        },
        index=pd.Index([4411, 4451, 5521, 5561, 9341], name="tour_id"),
    )
    df["TourMode"] = df["TourMode"].astype("category")
    df["time_period"] = df["time_period"].astype(
        pd.CategoricalDtype(categories=["AM", "MD"], ordered=False)
    )
    # time_period_alt is intentionally constructed backwards to test that bad ordering is handled
    df["time_period_alt"] = df["time_period_alt"].astype(
        pd.CategoricalDtype(categories=["MD", "AM"], ordered=False)
    )
    return construct(df)


@pytest.fixture
def skims_dataset() -> xr.Dataset:
    """Sample skims dataset with dummy data."""
    return xr.Dataset(
        {
            "cartime": xr.DataArray(
                [
                    [[11.1, 22.2, 33.3], [44.4, 55.5, 66.6], [77.7, 88.8, 99.9]],
                    [[1.1, 2.2, 3.3], [4.4, 5.5, 6.6], [7.7, 8.8, 9.9]],
                ],
                dims=["timeperiod", "otaz", "dtaz"],
                coords={
                    "timeperiod": ["AM", "MD"],
                    "otaz": [101, 102, 103],
                    "dtaz": [101, 102, 103],
                },
            )
        }
    )



from __future__ import annotations

import pandas as pd
from pytest import approx

import larch as lx
from larch import P, X


def test_swissmetro_101():
    m = lx.Model()
    m.title = "swissmetro example 01 (simple logit)"
    m.availability_co_vars = {
        1: "TRAIN_AV * (SP!=0)",
        2: "SM_AV",
        3: "CAR_AV * (SP!=0)",
    }
    m.choice_co_code = "CHOICE"
    m.utility_co[1] = P("ASC_TRAIN")
    m.utility_co[2] = 0
    m.utility_co[3] = P("ASC_CAR")
    m.utility_co[1] += X("TRAIN_TT") * P("B_TIME")
    m.utility_co[2] += X("SM_TT") * P("B_TIME")
    m.utility_co[3] += X("CAR_TT") * P("B_TIME")
    m.utility_co[1] += X("TRAIN_CO*(GA==0)") * P("B_COST")
    m.utility_co[2] += X("SM_CO*(GA==0)") * P("B_COST")
    m.utility_co[3] += X("CAR_CO") * P("B_COST")
    m.ordering = [("ASCs", "ASC.*"), ("LOS", "B_.*")]
    raw_data = pd.read_csv(lx.example_file("swissmetro.csv.gz")).rename_axis(
        index="CASEID"
    )
    raw_data.head()
    keep = raw_data.eval("PURPOSE in (1,3) and CHOICE != 0")
    selected_data = raw_data[keep]
    ds = lx.Dataset.construct.from_idco(
        selected_data, alts={1: "Train", 2: "SM", 3: "Car"}
    )
    m.datatree = ds
    m.set_cap(15)
    result = m.maximize_loglike(method="SLSQP")
    assert result.loglike == approx(-5331.252006971916)
    m.calculate_parameter_covariance()
    m.parameter_summary()
    assert m.pvals == approx([-0.15475, -0.70178, -0.010842, -0.012776], rel=1e-3)
    assert m.pstderr == approx(
        [0.0432357, 0.0548783, 0.000518335, 0.000568829], rel=1e-3
    )
    assert m.parameter_summary().data["t Stat"].values.astype(float) == approx(
        [-3.58, -12.79, -20.92, -22.46], rel=1e-2
    )




# Example code extracted automatically from documentation.
# Do not edit these files manually.
from ._017C_segmentation_males import example as _017C_segmentation_males
from ._017A_market_seg import example as _017A_market_seg
from ._016_cost_by_income import example as _016_cost_by_income
from ._003_zeroed_shared import example as _003_zeroed_shared
from ._017D_segmentation_female import example as _017D_segmentation_female
from ._013_cbd import example as _013_cbd
from ._017_mnl import example as _017_mnl
from ._014_work_zone_density import example as _014_work_zone_density
from ._022_nl import example as _022_nl
from ._007_ovtt_by_distance import example as _007_ovtt_by_distance
from ._017B_alt_segmentation import example as _017B_alt_segmentation
from ._012_autos_per_adult import example as _012_autos_per_adult
from ._009_ttr_4 import example as _009_ttr_4
from ._030_mnl_constrain import example as _030_mnl_constrain
from ._006_motorized_times import example as _006_motorized_times
from ._015_cbd_and_work_density import example as _015_cbd_and_work_density
from ._011_veh_per_worker import example as _011_veh_per_worker
from ._005_motorized_travel_time import example as _005_motorized_travel_time
from ._010_auto_per_hh import example as _010_auto_per_hh
from ._008_ttr import example as _008_ttr
from ._002_motorized import example as _002_motorized
from ._001_mnl import example as _001_mnl
from ._102_swissmetro_weighted import example as _102_swissmetro_weighted
from ._151_swissmetro_panel_latent_class import example as _151_swissmetro_panel_latent_class
from ._109_swissmetro_nl import example as _109_swissmetro_nl
from ._101_swissmetro_mnl import example as _101_swissmetro_mnl
from ._200_exampville import example as _200_exampville
from ._251_bySpec import example as _251_bySpec
from ._202_exville_mc_logsums import example as _202_exville_mc_logsums
from ._203_exville_dest import example as _203_exville_dest
from ._201_exville_mode_choice import example as _201_exville_mode_choice
from ._204_exville_dest_only import example as _204_exville_dest_only
from ._xlogit_test import example as _xlogit_test

ex = {}
ex['017C'] = _017C_segmentation_males
ex['017A'] = _017A_market_seg
ex[16] = _016_cost_by_income
ex[3] = _003_zeroed_shared
ex['017D'] = _017D_segmentation_female
ex[13] = _013_cbd
ex[17] = _017_mnl
ex[14] = _014_work_zone_density
ex[22] = _022_nl
ex[7] = _007_ovtt_by_distance
ex['017B'] = _017B_alt_segmentation
ex[12] = _012_autos_per_adult
ex[9] = _009_ttr_4
ex[30] = _030_mnl_constrain
ex[6] = _006_motorized_times
ex[15] = _015_cbd_and_work_density
ex[11] = _011_veh_per_worker
ex[5] = _005_motorized_travel_time
ex[10] = _010_auto_per_hh
ex[8] = _008_ttr
ex[2] = _002_motorized
ex[1] = _001_mnl
ex[102] = _102_swissmetro_weighted
ex[151] = _151_swissmetro_panel_latent_class
ex[109] = _109_swissmetro_nl
ex[101] = _101_swissmetro_mnl
ex[200] = _200_exampville
ex[251] = _251_bySpec
ex[202] = _202_exville_mc_logsums
ex[203] = _203_exville_dest
ex[201] = _201_exville_mode_choice
ex[204] = _204_exville_dest_only
ex['xlogit'] = _xlogit_test



from collections.abc import Mapping

import numba as nb
import numpy as np


@nb.njit(error_model="numpy", fastmath=True, cache=True)
def _utility_to_probability(
    n_alts,
    edges_up,  # int input shape=[edges]
    edges_dn,  # int input shape=[edges]
    mu_params,  # float input shape=[nests]
    start_slots,  # int input shape=[nests]
    len_slots,  # int input shape=[nests]
    only_utility,  #
    utility,  # float output shape=[nodes]
    logprob,  # float output shape=[nodes]
    probability,  # float output shape=[nodes]
):
    for up in range(n_alts, utility.size):
        up_nest = up - n_alts
        n_children_for_parent = len_slots[up_nest]
        shifter = -np.inf
        shifter_position = -1
        mu_up = mu_params[up_nest]
        if mu_up:
            for n in range(n_children_for_parent):
                edge = start_slots[up_nest] + n
                dn = edges_dn[edge]
                if utility[dn] > -np.inf:
                    z = utility[dn] / mu_up
                    if z > shifter:
                        shifter = z
                        shifter_position = dn
            for n in range(n_children_for_parent):
                edge = start_slots[up_nest] + n
                dn = edges_dn[edge]
                if utility[dn] > -np.inf:
                    if shifter_position == dn:
                        utility[up] += 1
                    else:
                        utility[up] += np.exp((utility[dn] / mu_up) - shifter)
            utility[up] = (np.log(utility[up]) + shifter) * mu_up
        else:  # mu_up is zero
            for n in range(n_children_for_parent):
                edge = start_slots[up_nest] + n
                dn = edges_dn[edge]
                if utility[dn] > utility[up]:
                    utility[up] = utility[dn]

    if only_utility:
        return

    for s in range(edges_up.size):
        dn = edges_dn[s]
        up = edges_up[s]
        mu_up = mu_params[up - n_alts]
        if np.isinf(utility[up]) and utility[up] < 0:
            logprob[dn] = -np.inf
        else:
            logprob[dn] = (utility[dn] - utility[up]) / mu_up

    # logprob becomes conditional_probability
    conditional_probability = logprob
    for i in range(logprob.size):
        conditional_probability[i] = np.exp(logprob[i])

    # probability
    probability[-1] = 1.0
    for s in range(edges_up.size - 1, -1, -1):
        dn = edges_dn[s]
        up = edges_up[s]
        probability[dn] = probability[up] * conditional_probability[dn]


@nb.njit
def _utility_to_probability_array(
    edges_up,  # int input shape=[edges]
    edges_dn,  # int input shape=[edges]
    mu_params,  # float input shape=[nests]
    start_slots,  # int input shape=[nests]
    len_slots,  # int input shape=[nests]
    utility,  # float output shape=[nodes]
):
    pr = np.zeros_like(utility)
    n_nodes = mu_params.size
    n_alts = utility.shape[1]
    for n in range(utility.shape[0]):
        _pr = np.zeros(n_nodes, dtype=np.float32)
        _util = np.zeros(n_nodes, dtype=np.float32)
        _logpr = np.zeros(n_nodes, dtype=np.float32)
        _util[:n_alts] = utility[n, :n_alts]
        _utility_to_probability(
            n_alts,
            edges_up,  # int input shape=[edges]
            edges_dn,  # int input shape=[edges]
            mu_params[n_alts:],  # float input shape=[nests]
            start_slots[n_alts:],  # int input shape=[nests]
            len_slots[n_alts:],  # int input shape=[nests]
            False,  #
            _util,  # float output shape=[nodes]
            _logpr,  # float output shape=[nodes]
            _pr,  # float output shape=[nodes]
        )
        pr[n, :] = _pr[:n_alts]
    return pr


@nb.njit
def _utility_to_logsums_array(
    edges_up,  # int input shape=[edges]
    edges_dn,  # int input shape=[edges]
    mu_params,  # float input shape=[nests]
    start_slots,  # int input shape=[nests]
    len_slots,  # int input shape=[nests]
    utility,  # float output shape=[nodes]
):
    logsums = np.zeros(utility.shape[0], dtype=np.float32)
    n_nodes = mu_params.size
    n_alts = utility.shape[1]
    for n in range(utility.shape[0]):
        _pr = np.zeros(n_nodes, dtype=np.float32)
        _util = np.zeros(n_nodes, dtype=np.float32)
        _logpr = np.zeros(n_nodes, dtype=np.float32)
        _util[:n_alts] = utility[n, :n_alts]
        _utility_to_probability(
            n_alts,
            edges_up,  # int input shape=[edges]
            edges_dn,  # int input shape=[edges]
            mu_params[n_alts:],  # float input shape=[nests]
            start_slots[n_alts:],  # int input shape=[nests]
            len_slots[n_alts:],  # int input shape=[nests]
            True,  #
            _util,  # float output shape=[nodes]
            _logpr,  # float output shape=[nodes]
            _pr,  # float output shape=[nodes]
        )
        logsums[n] = _util[-1]
    return logsums


def construct_nesting_tree(alternatives, nesting_settings):
    """
    Construct a larch NestingTree from ActivitySim settings.

    Parameters
    ----------
    alternatives : Mapping or Sequence
        If given as a Mapping (dict), the keys are the alternative names
        as strings, and the values are alternative code numbers to use
        in larch.  If given as a Sequence, the values are the alternative
        names, and unique sequential codes will be created starting from 1.
    nesting_settings : Mapping
        The 'NESTS' section of the ActivitySim config file.

    Returns
    -------
    NestingTree
    """
    try:
        from larch.model.tree import NestingTree
    except ImportError:
        raise ImportError("larch is required to construct nesting trees") from None

    if not isinstance(alternatives, Mapping):
        alt_names = list(alternatives)
        alt_codes = np.arange(1, len(alt_names) + 1)
        alternatives = dict(zip(alt_names, alt_codes))

    tree = NestingTree()
    nest_names_to_codes = alternatives.copy()
    nest_names_to_codes["root"] = 0
    for alt_name, alt_code in alternatives.items():
        tree.add_node(alt_code, name=alt_name)

    def make_nest(cfg, parent_code=0):
        nonlocal nest_names_to_codes
        if cfg["name"] != "root":
            if cfg["name"] not in nest_names_to_codes:
                n = tree.new_node(
                    name=cfg["name"],
                    parameter=str(cfg["coefficient"]),
                    parent=parent_code,
                )
                nest_names_to_codes[cfg["name"]] = n
            else:
                tree.add_edge(parent_code, nest_names_to_codes[cfg["name"]])
        for a in cfg["alternatives"]:
            if isinstance(a, str):
                tree.add_edge(nest_names_to_codes[cfg["name"]], nest_names_to_codes[a])
            else:
                make_nest(a, parent_code=nest_names_to_codes[cfg["name"]])

    make_nest(nesting_settings)

    return tree
